from typing import Dict, Optional, Set, Tuple, Iterator, Union
import archinfo

import claripy

from angr.storage.memory_object import bv_slice


class MultiValues:
    """
    Represents a byte vector where each byte can have one or multiple values.

    As an implementation optimization (so that we do not create excessive sets and dicts), self._single_value stores a
    claripy AST when this MultiValues object represents only one value at offset 0.
    """

    __slots__ = (
        "_values",
        "_single_value",
    )

    _single_value: Optional[claripy.ast.Bits]
    _values: Optional[Dict[int, Set[claripy.ast.Bits]]]

    def __init__(
        self,
        v: Union[claripy.ast.Bits, "MultiValues", None, Dict[int, Set[claripy.ast.Bits]]] = None,
        offset_to_values=None,
    ):
        if v is not None and offset_to_values is not None:
            raise TypeError("You cannot specify v and offset_to_values at the same time!")

        self._single_value = (
            None
            if v is None
            else v if isinstance(v, claripy.ast.Bits) else v._single_value if isinstance(v, MultiValues) else None
        )
        self._values = (
            offset_to_values
            if offset_to_values is not None
            else v if isinstance(v, dict) else v._values if isinstance(v, MultiValues) else None
        )

        if self._single_value is None and self._values is None:
            self._values = {}

        # if only one value is passed in, assign it to self._single_value
        if self._values:
            if len(self._values) == 1 and 0 in self._values and len(self._values[0]) == 1:
                self._single_value = next(iter(self._values[0]))
                self._values = None

        if self._values:
            # sanity check
            for vs in self._values.values():
                if not isinstance(vs, set):
                    raise TypeError("Each value in offset_to_values must be a set!")

    def add_value(self, offset: int, value: claripy.ast.Bits) -> None:
        if len(value) == 0:
            return
        if self._single_value is not None:
            if len(self._single_value) == 0:
                if offset == 0:
                    self._single_value = value
                else:
                    self._single_value = None
                    self._values = {offset: {value}}
                return
            self._values = {0: {self._single_value}}
            self._single_value = None

        if self._values is None:
            self._values = {}

        if offset not in self._values:
            self._values[offset] = set()

        # if value overlaps with existing values, we need to break them down

        succ_offset = self._adjacent_offset(offset, before=False)
        if succ_offset is not None:
            value_end = offset + value.length // 8
            if value_end > succ_offset:
                # value is too long. we need to break value into two
                mid_value_size = succ_offset - offset
                remaining_value = value[value.length - mid_value_size * 8 - 1 : 0]
                # update value
                value = value[value.length - 1 : value.length - mid_value_size * 8]
                self.add_value(succ_offset, remaining_value)

        if self._values[offset]:
            curr_value_size = next(iter(self._values[offset])).length // 8
            if curr_value_size > value.length // 8:
                # we need to break existing values
                new_curr_values = set()
                remaining_values = set()
                for v in self._values[offset]:
                    new_curr_values.add(v[v.length - 1 : v.length - value.length])
                    remaining_values.add(v[v.length - value.length - 1 : 0])
                self._values[offset] = new_curr_values
                for v in remaining_values:
                    self.add_value(offset + value.length // 8, v)
            elif curr_value_size < value.length // 8:
                # value is too long. we need to break value into two
                remaining_value = value[value.length - curr_value_size * 8 - 1 : 0]
                # update value
                value = value[value.length - 1 : value.length - curr_value_size * 8]
                self.add_value(offset + curr_value_size, remaining_value)

        self._values[offset].add(value)

        pre_offset = self._adjacent_offset(offset, before=True)
        if pre_offset is not None:
            pre_values = self._values[pre_offset]
            pre_values_size = next(iter(pre_values)).length // 8
            if pre_offset + pre_values_size > offset:
                # we need to break the preceding values
                new_pre_value_size = offset - pre_offset
                new_pre_values = set()
                remaining_values = set()
                for v in pre_values:
                    new_pre_values.add(v[v.length - 1 : v.length - new_pre_value_size * 8])
                    remaining_values.add(v[v.length - new_pre_value_size * 8 - 1 : 0])
                self._values[pre_offset] = new_pre_values
                for v in remaining_values:
                    self.add_value(offset, v)

    def one_value(self, strip_annotations: bool = False) -> Optional[claripy.ast.Bits]:
        if self._single_value is not None:
            return self._single_value

        assert self._values is not None

        if len(self._values) == 1 and len(self._values[0]) == 1:
            return next(iter(self._values[0]))
        if strip_annotations:
            all_values_wo_annotations = {
                bv.remove_annotations(bv.annotations) if bv.annotations else bv for bv in self._values[0]
            }
            if len(all_values_wo_annotations) == 1:
                return next(iter(all_values_wo_annotations))
        return None

    def __len__(self) -> int:
        if self._single_value is not None:
            return self._single_value.length

        assert self._values is not None

        if len(self._values) == 0:
            return 0

        max_offset = max(self._values.keys())
        max_len = max(x.size() for x in self._values[max_offset])
        return max_offset * 8 + max_len  # FIXME: we are assuming byte_width of 8

    def merge(self, mv: "MultiValues") -> "MultiValues":
        new_values = {k: set(v) for k, v in self.items()}
        for off, vs in mv.items():
            if off not in new_values:
                new_values[off] = set(vs)
            else:
                new_values[off] |= vs
        return MultiValues(offset_to_values=new_values)

    def __eq__(self, other) -> bool:
        if not isinstance(other, MultiValues):
            return False
        if self._single_value is not None and other._single_value is not None:
            return self._single_value is other._single_value
        if self._single_value is not None and other._single_value is None:
            return False
        if self._single_value is None and other._single_value is not None:
            return False
        assert self._values is not None
        assert other._values is not None
        if set(self._values.keys()) != set(other._values.keys()):
            return False
        for k in self._values.keys():
            if self._values[k] != other._values[k]:
                return False
        return True

    def __repr__(self):
        if self._single_value is not None:
            return f"<{self.__class__.__name__}({self._single_value})>"
        return f"<{self.__class__.__name__}({self._values})>"

    def __contains__(self, offset: int) -> bool:
        if self._single_value is not None:
            return offset == 0
        return False if not self._values else offset in self._values

    def __getitem__(self, offset: int) -> Set[claripy.ast.Bits]:
        if self._single_value is not None:
            if offset == 0:
                return {self._single_value}
            raise KeyError()
        elif not self._values:
            raise KeyError()
        return self._values[offset]

    def keys(self) -> Set[int]:
        if self._single_value is not None:
            return {0}
        return set() if not self._values else set(self._values.keys())

    def values(self) -> Iterator[Set[claripy.ast.Bits]]:
        if self._single_value is not None:
            yield {self._single_value}
        else:
            if self._values is None:
                return
            yield from self._values.values()

    def items(self) -> Iterator[Tuple[int, Set[claripy.ast.Bits]]]:
        if self._single_value is not None:
            yield 0, {self._single_value}
        else:
            if self._values is None:
                yield 0, set()
            else:
                yield from self._values.items()

    def count(self) -> int:
        if self._single_value is not None:
            return 1
        if self._values is None:
            return 0
        return len(self._values)

    def extract(self, offset: int, length: int, endness: str) -> "MultiValues":
        end = offset + length
        result = MultiValues(claripy.BVV(b""))
        for obj_offset, values in self.items():
            if obj_offset >= end:
                break
            for value in values:
                obj_length = len(value) // 8
                if obj_offset + obj_length < offset:
                    continue

                slice_start = max(0, offset - obj_offset)
                slice_end = min(obj_length, end - obj_offset)
                sliced = bv_slice(value, slice_start, slice_end - slice_start, endness == archinfo.Endness.LE, 8)
                if len(sliced):
                    result.add_value(max(0, obj_offset - offset), sliced)

        return result

    def concat(self, other: Union["MultiValues", claripy.ast.Bits, bytes]) -> "MultiValues":
        if isinstance(other, bytes):
            other = claripy.BVV(other)
        if isinstance(other, claripy.ast.Bits):
            other = MultiValues(other)
        offset = len(self) // 8
        result = MultiValues(self)
        for k, v in other.items():
            for v2 in v:
                result.add_value(k + offset, v2)
        return result

    #
    # Private methods
    #

    def _adjacent_offset(self, offset: int, before: bool = True) -> Optional[int]:
        """
        Find the offset that is right before or after the given offset.

        :param offset:  The specified offset.
        :param before:  True if we want to find the offset right before the specified offset, False if we want to find
                        the offset right after the specified offset.
        :return:        The adjacent offset as requested. If the requested adjacent offset does not exist, return None.
        """

        sorted_offsets = list(sorted(self._values.keys())) if self._values is not None else [0]

        for i, off in enumerate(sorted_offsets):
            if off == offset:
                if before:
                    return sorted_offsets[i - 1] if i > 0 else None
                else:
                    return sorted_offsets[i + 1] if i + 1 < len(sorted_offsets) else None
            if off > offset:
                # we missed it...
                return None
        return None



from ailment.expression import UnaryOp, BinaryOp

from .base import PeepholeOptimizationExprBase


class InvertNegatedLogicalConjunctionsAndDisjunctions(PeepholeOptimizationExprBase):
    """
    Push negations into subexpressions of logical conjunctions and logical disjunctions.
    """

    __slots__ = ()

    NAME = "!(A && B) => A || B; !(A || B) => A && B"
    expr_classes = (UnaryOp,)  # all expressions are allowed

    def optimize(self, expr: UnaryOp, **kwargs):
        if expr.op == "Not":
            if isinstance(expr.operand, BinaryOp):
                if expr.operand.op == "LogicalAnd":
                    inner_operands = [
                        UnaryOp(None, "Not", expr.operand.operands[0], **expr.operand.operands[0].tags),
                        UnaryOp(None, "Not", expr.operand.operands[1], **expr.operand.operands[1].tags),
                    ]
                    return BinaryOp(
                        expr.operand.idx,
                        "LogicalOr",
                        inner_operands,
                        expr.operand.signed,
                        variable=expr.operand.variable,
                        variable_offset=expr.operand.variable_offset,
                        bits=expr.operand.bits,
                        **expr.tags,
                    )
                elif expr.operand.op == "LogicalOr":
                    inner_operands = [
                        UnaryOp(None, "Not", expr.operand.operands[0], **expr.operand.operands[0].tags),
                        UnaryOp(None, "Not", expr.operand.operands[1], **expr.operand.operands[1].tags),
                    ]
                    return BinaryOp(
                        expr.idx,
                        "LogicalAnd",
                        inner_operands,
                        expr.operand.signed,
                        variable=expr.operand.variable,
                        variable_offset=expr.operand.variable_offset,
                        bits=expr.operand.bits,
                        **expr.tags,
                    )

        return None



from typing import Optional

import claripy

from ..errors import SimMemoryError


def obj_bit_size(o):
    if type(o) is bytes:
        return len(o) * 8
    return o.size()


# TODO: get rid of is_bytes and have the bytes-backed objects be a separate class


class SimMemoryObject:
    """
    A SimMemoryObject is a reference to a byte or several bytes in a specific object in memory. It should be used only
    by the bottom layer of memory.
    """

    __slots__ = (
        "is_bytes",
        "_byte_width",
        "base",
        "object",
        "length",
        "endness",
        "_concrete_bytes",
    )

    def __init__(self, obj, base, endness, length=None, byte_width=8):
        if type(obj) is bytes:
            assert byte_width == 8

        elif not isinstance(obj, claripy.ast.Base):
            raise SimMemoryError("memory can only store claripy Expression")

        self.is_bytes = type(obj) is bytes
        if self.is_bytes and endness != "Iend_BE":
            raise SimMemoryError("bytes can only be stored big-endian")
        self._byte_width = byte_width
        self.base = base
        self.object = obj
        self.length = obj_bit_size(obj) // self._byte_width if length is None else length
        self.endness = endness
        self._concrete_bytes: Optional[bytes] = None

    def size(self):
        return self.length * self._byte_width

    def __len__(self):
        return self.size()

    @property
    def variables(self):
        return self.object.variables

    @property
    def cache_key(self):
        return self.object.cache_key

    @property
    def symbolic(self):
        return self.object.symbolic

    @property
    def last_addr(self):
        return self.base + self.length - 1

    def concrete_bytes(self, offset: int, size: int) -> Optional[bytes]:
        if self._concrete_bytes is None:
            if isinstance(self.object, claripy.ast.Bits) and self.object.op == "BVV" and not self.object.annotations:
                self._concrete_bytes = self.object.concrete_value.to_bytes(len(self.object) // self._byte_width, "big")

        if self._concrete_bytes is None:
            return None

        return self._concrete_bytes[offset : offset + size]

    def includes(self, x):
        return 0 <= x - self.base < self.length

    def bytes_at(self, addr, length, allow_concrete=False, endness="Iend_BE"):
        rev = endness != self.endness
        if allow_concrete and rev:
            raise Exception("allow_concrete must be used with the stored endness")

        if self.is_bytes:
            if addr == self.base and length == self.length:
                o = self.object
            else:
                start = addr - self.base
                end = start + length
                o = self.object[start:end]

            return o if allow_concrete else claripy.BVV(o)

        else:
            offset = addr - self.base
            try:
                thing = bv_slice(self.object, offset, length, self.endness == "Iend_LE", self._byte_width)
            except claripy.ClaripyOperationError:
                # hacks to handle address space wrapping
                if offset >= 0:
                    raise
                if offset + 2**32 >= 0:
                    offset += 2**32
                elif offset + 2**64 >= 0:
                    offset += 2**64
                else:
                    raise
                thing = bv_slice(self.object, offset, length, self.endness == "Iend_LE", self._byte_width)

            if self.endness != endness:
                thing = thing.reversed
            return thing

    def _object_equals(self, other):
        if self.is_bytes != other.is_bytes:
            return False

        if self.is_bytes:
            return self.object == other.object
        else:
            return self.object.cache_key == other.object.cache_key

    def _length_equals(self, other):
        if type(self.length) is not type(other.length):  # noqa: E721
            return False

        if isinstance(self.length, int):
            return self.length == other.length
        else:
            return self.length.cache_key == other.length.cache_key

    def __eq__(self, other):
        if self is other:
            return True

        if type(other) is not SimMemoryObject:
            return NotImplemented

        return self.base == other.base and self._object_equals(other) and self._length_equals(other)

    def __hash__(self):
        obj_hash = hash(self.object) if self.is_bytes else self.object.cache_key
        return hash((obj_hash, self.base, hash(self.length)))

    def __ne__(self, other):
        return not self == other

    def __repr__(self):
        return "MO(%s)" % self.object


class SimLabeledMemoryObject(SimMemoryObject):
    __slots__ = ("label",)

    def __init__(self, obj, base, endness, length=None, byte_width=8, label=None):
        super().__init__(obj, base, endness, length=length, byte_width=byte_width)
        self.label = label


def bv_slice(value: claripy.ast.BV, offset: int, size: int, rev: bool, bw: int) -> claripy.ast.BV:
    """
    Extremely cute utility to pretend you've serialized a value to stored bytes, sliced it a la python slicing, and then
    deserialized those bytes to an integer again.

    :param value:   The bitvector to slice
    :param offset:  The byte offset from the first stored byte to slice from, or a negative offset from the end.
    :param size:    The number of bytes to return. If None, return all bytes from the offset to the end. If larger than
                    the number of bytes from the offset to the end, return all bytes from the offset to the end.
    :param rev:     Whether the pretend-serialization should be little-endian
    :param bw:      The byte width
    :return:        The new bitvector
    """

    vsize = len(value) // bw

    if offset < 0:
        offset = vsize + offset

    if size is None or offset + size > vsize:
        size = vsize - offset
    if rev:
        offset = vsize - (offset + size)

    if offset == 0 and size == vsize:
        return value

    bitstart = len(value) - offset * bw
    if size == 0:
        return claripy.BVV(b"")
    return value[bitstart - 1 : bitstart - size * bw]



import angr
import archinfo
import logging

l = logging.getLogger(name=__name__)


class sigaction(angr.SimProcedure):
    def run(self, signum, act, oldact):
        l.warning("Calling sigaction for signal %s - this is emulated as a nop", signum)
        if not self.state.solver.is_true(oldact == 0):
            if isinstance(self.arch, archinfo.ArchAMD64):
                self.state.memory.store(
                    oldact, self.state.solver.BVS("sigaction_oldact", 152 * 8, key=("api", "sigaction", "oldact"))
                )
            elif isinstance(self.arch, archinfo.ArchX86):
                self.state.memory.store(
                    oldact, self.state.solver.BVS("sigaction_oldact", 140 * 8, key=("api", "sigaction", "oldact"))
                )
        return 0



# pylint:disable=line-too-long
import logging

from ...sim_type import SimTypeFunction,     SimTypeShort, SimTypeInt, SimTypeLong, SimTypeLongLong, SimTypeDouble, SimTypeFloat,     SimTypePointer,     SimTypeChar,     SimStruct,     SimTypeFixedSizeArray,     SimTypeBottom,     SimUnion,     SimTypeBool
from ...calling_conventions import SimCCStdcall, SimCCMicrosoftAMD64
from .. import SIM_PROCEDURES as P
from . import SimLibrary


_l = logging.getLogger(name=__name__)


lib = SimLibrary()
lib.set_default_cc('X86', SimCCStdcall)
lib.set_default_cc('AMD64', SimCCMicrosoftAMD64)
lib.set_library_names("d3d11.dll")
prototypes = \
    {
        #
        'D3D11CreateDevice': SimTypeFunction([SimTypeBottom(label="IDXGIAdapter"), SimTypeInt(signed=False, label="D3D_DRIVER_TYPE"), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="D3D11_CREATE_DEVICE_FLAG"), SimTypePointer(SimTypeInt(signed=False, label="D3D_FEATURE_LEVEL"), label="LPArray", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="ID3D11Device"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="D3D_FEATURE_LEVEL"), offset=0), SimTypePointer(SimTypeBottom(label="ID3D11DeviceContext"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pAdapter", "DriverType", "Software", "Flags", "pFeatureLevels", "FeatureLevels", "SDKVersion", "ppDevice", "pFeatureLevel", "ppImmediateContext"]),
        #
        'D3D11CreateDeviceAndSwapChain': SimTypeFunction([SimTypeBottom(label="IDXGIAdapter"), SimTypeInt(signed=False, label="D3D_DRIVER_TYPE"), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="D3D11_CREATE_DEVICE_FLAG"), SimTypePointer(SimTypeInt(signed=False, label="D3D_FEATURE_LEVEL"), label="LPArray", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"BufferDesc": SimStruct({"Width": SimTypeInt(signed=False, label="UInt32"), "Height": SimTypeInt(signed=False, label="UInt32"), "RefreshRate": SimStruct({"Numerator": SimTypeInt(signed=False, label="UInt32"), "Denominator": SimTypeInt(signed=False, label="UInt32")}, name="DXGI_RATIONAL", pack=False, align=None), "Format": SimTypeInt(signed=False, label="DXGI_FORMAT"), "ScanlineOrdering": SimTypeInt(signed=False, label="DXGI_MODE_SCANLINE_ORDER"), "Scaling": SimTypeInt(signed=False, label="DXGI_MODE_SCALING")}, name="DXGI_MODE_DESC", pack=False, align=None), "SampleDesc": SimStruct({"Count": SimTypeInt(signed=False, label="UInt32"), "Quality": SimTypeInt(signed=False, label="UInt32")}, name="DXGI_SAMPLE_DESC", pack=False, align=None), "BufferUsage": SimTypeInt(signed=False, label="UInt32"), "BufferCount": SimTypeInt(signed=False, label="UInt32"), "OutputWindow": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "Windowed": SimTypeInt(signed=True, label="Int32"), "SwapEffect": SimTypeInt(signed=False, label="DXGI_SWAP_EFFECT"), "Flags": SimTypeInt(signed=False, label="UInt32")}, name="DXGI_SWAP_CHAIN_DESC", pack=False, align=None), offset=0), SimTypePointer(SimTypeBottom(label="IDXGISwapChain"), offset=0), SimTypePointer(SimTypeBottom(label="ID3D11Device"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="D3D_FEATURE_LEVEL"), offset=0), SimTypePointer(SimTypeBottom(label="ID3D11DeviceContext"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pAdapter", "DriverType", "Software", "Flags", "pFeatureLevels", "FeatureLevels", "SDKVersion", "pSwapChainDesc", "ppSwapChain", "ppDevice", "pFeatureLevel", "ppImmediateContext"]),
        #
        'D3D11On12CreateDevice': SimTypeFunction([SimTypeBottom(label="IUnknown"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="D3D_FEATURE_LEVEL"), label="LPArray", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="IUnknown"), label="LPArray", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="ID3D11Device"), offset=0), SimTypePointer(SimTypeBottom(label="ID3D11DeviceContext"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="D3D_FEATURE_LEVEL"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pDevice", "Flags", "pFeatureLevels", "FeatureLevels", "ppCommandQueues", "NumQueues", "NodeMask", "ppDevice", "ppImmediateContext", "pChosenFeatureLevel"]),
        #
        'CreateDirect3D11DeviceFromDXGIDevice': SimTypeFunction([SimTypeBottom(label="IDXGIDevice"), SimTypePointer(SimTypeBottom(label="IInspectable"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["dxgiDevice", "graphicsDevice"]),
        #
        'CreateDirect3D11SurfaceFromDXGISurface': SimTypeFunction([SimTypeBottom(label="IDXGISurface"), SimTypePointer(SimTypeBottom(label="IInspectable"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["dgxiSurface", "graphicsSurface"]),
    }

lib.set_prototypes(prototypes)



import functools
import itertools
import contextlib
import weakref

import logging

from typing import Type, TypeVar, TYPE_CHECKING

from archinfo import Arch

l = logging.getLogger(name=__name__)

import claripy
import archinfo
from archinfo.arch_soot import SootAddressDescriptor

from .misc.plugins import PluginHub, PluginPreset
from .sim_state_options import SimStateOptions
from .state_plugins import SimStatePlugin


def arch_overrideable(f):
    @functools.wraps(f)
    def wrapped_f(self, *args, **kwargs):
        if hasattr(self.arch, f.__name__):
            arch_f = getattr(self.arch, f.__name__)
            return arch_f(self, *args, **kwargs)
        else:
            return f(self, *args, **kwargs)

    return wrapped_f


# This is a counter for the state-merging symbolic variables
merge_counter = itertools.count()

_complained_se = False


# pylint: disable=not-callable
class SimState(PluginHub):
    """
    The SimState represents the state of a program, including its memory, registers, and so forth.

    :param angr.Project project:    The project instance.
    :param archinfo.Arch|str arch:  The architecture of the state.

    :ivar regs:         A convenient view of the state's registers, where each register is a property
    :ivar mem:          A convenient view of the state's memory, a :class:`angr.state_plugins.view.SimMemView`
    :ivar registers:    The state's register file as a flat memory region
    :ivar memory:       The state's memory as a flat memory region
    :ivar solver:       The symbolic solver and variable manager for this state
    :ivar inspect:      The breakpoint manager, a :class:`angr.state_plugins.inspect.SimInspector`
    :ivar log:          Information about the state's history
    :ivar scratch:      Information about the current execution step
    :ivar posix:        MISNOMER: information about the operating system or environment model
    :ivar fs:           The current state of the simulated filesystem
    :ivar libc:         Information about the standard library we are emulating
    :ivar cgc:          Information about the cgc environment
    :ivar uc_manager:   Control of under-constrained symbolic execution
    :ivar unicorn:      Control of the Unicorn Engine
    """

    # Type Annotations for default plugins to allow type inference
    solver: "SimSolver"
    posix: "SimSystemPosix"
    registers: "DefaultMemory"
    regs: "SimRegNameView"
    memory: "DefaultMemory"
    callstack: "CallStack"
    mem: "SimMemView"
    history: "SimStateHistory"
    inspect: "SimInspector"
    jni_references: "SimStateJNIReferences"
    scratch: "SimStateScratch"

    def __init__(
        self,
        project=None,
        arch=None,
        plugins=None,
        mode=None,
        options=None,
        add_options=None,
        remove_options=None,
        special_memory_filler=None,
        os_name=None,
        plugin_preset="default",
        cle_memory_backer=None,
        dict_memory_backer=None,
        permissions_map=None,
        default_permissions=3,
        stack_perms=None,
        stack_end=None,
        stack_size=None,
        regioned_memory_cls=None,
        **kwargs,
    ):
        if kwargs:
            l.warning("Unused keyword arguments passed to SimState: %s", " ".join(kwargs))
        super().__init__()
        self.project = project

        # Java & Java JNI
        self._is_java_project = self.project and self.project.is_java_project
        self._is_java_jni_project = self.project and self.project.is_java_jni_project

        # Arch
        if self._is_java_jni_project:
            self._arch = {"soot": project.arch, "vex": project.simos.native_simos.arch}
            # This flag indicates whether the current ip is a native address or
            # a soot address descriptor.
            # Note: We cannot solely rely on the ip to make that decsision,
            #       because the registers (storing the ip) are part of the
            #       plugins that are getting toggled (=> mutual dependence).
            self.ip_is_soot_addr = False
        else:
            self._arch = arch if arch is not None else project.arch.copy() if project is not None else None
            if type(self._arch) is str:
                self._arch = archinfo.arch_from_id(self._arch)

        # the options
        if options is None:
            if mode is None:
                l.warning("SimState defaulting to symbolic mode.")
                mode = "symbolic"
            options = o.modes[mode]

        if isinstance(options, (set, list)):
            options = SimStateOptions(options)
        if add_options is not None:
            options |= add_options
        if remove_options is not None:
            options -= remove_options
        self.options = options
        self.mode = mode
        self.supports_inspect = False

        # OS name
        self.os_name = os_name

        # This is used in static mode as we don't have any constraints there
        self._satisfiable = True

        self.uninitialized_access_handler = None
        self._special_memory_filler = special_memory_filler

        # this is a global condition, applied to all added constraints, memory reads, etc
        self._global_condition = None
        self.ip_constraints = []

        # plugins. lord help us
        if plugin_preset is not None:
            self.use_plugin_preset(plugin_preset)

        if plugins is not None:
            for n, p in plugins.items():
                self.register_plugin(n, p, inhibit_init=True)

        if not self.has_plugin("memory"):
            # We don't set the memory endness because, unlike registers, it's hard to understand
            # which endness the data should be read.

            # If they didn't provide us with either a memory plugin or a plugin preset to use,
            # we have no choice but to use the 'default' plugin preset.
            if self.plugin_preset is None:
                self.use_plugin_preset("default")

            # Determine memory backend
            if self._is_java_project and not self._is_java_jni_project:
                sim_memory_cls = self.plugin_preset.request_plugin("javavm_memory")
                sim_memory = sim_memory_cls(memory_id="mem")

            elif o.ABSTRACT_MEMORY in self.options:
                # We use SimAbstractMemory in static mode.
                # Convert memory_backer into 'global' region.
                if cle_memory_backer is not None:
                    cle_memory_backer = {"global": cle_memory_backer}
                if dict_memory_backer is not None:
                    dict_memory_backer = {"global": dict_memory_backer}

                # TODO: support permissions backer in SimAbstractMemory
                sim_memory_cls = self.plugin_preset.request_plugin("abs_memory")
                sim_memory = sim_memory_cls(
                    cle_memory_backer=cle_memory_backer,
                    dict_memory_backer=dict_memory_backer,
                    memory_id="mem",
                    regioned_memory_cls=regioned_memory_cls,
                )

            elif o.FAST_MEMORY in self.options:
                sim_memory_cls = self.plugin_preset.request_plugin("fast_memory")
                sim_memory = sim_memory_cls(memory_id="mem")

            else:
                sim_memory_cls = self.plugin_preset.request_plugin("sym_memory")
                sim_memory = sim_memory_cls(
                    cle_memory_backer=cle_memory_backer,
                    dict_memory_backer=dict_memory_backer,
                    memory_id="mem",
                    permissions_map=permissions_map,
                    default_permissions=default_permissions,
                    stack_perms=stack_perms,
                    stack_end=stack_end,
                    stack_size=stack_size,
                )

            # Add memory plugin
            if not self._is_java_jni_project:
                self.register_plugin("memory", sim_memory, inhibit_init=True)

            else:
                # In case of the JavaVM with JNI support, we add two `memory` plugins; one for modeling the
                # native memory and another one for the JavaVM memory.
                native_sim_memory = sim_memory
                javavm_sim_memory_cls = self.plugin_preset.request_plugin("javavm_memory")
                javavm_sim_memory = javavm_sim_memory_cls(memory_id="mem")
                self.register_plugin("memory_soot", javavm_sim_memory, inhibit_init=True)
                self.register_plugin("memory_vex", native_sim_memory, inhibit_init=True)

        if not self.has_plugin("registers"):
            # Same as for 'memory' plugin.
            if self.plugin_preset is None:
                self.use_plugin_preset("default")

            # Get register endness
            if self._is_java_jni_project:
                register_endness = self._arch["vex"].register_endness
            else:
                register_endness = self.arch.register_endness

            # Determine register backend
            if self._is_java_project and not self._is_java_jni_project:
                sim_registers_cls = self.plugin_preset.request_plugin("keyvalue_memory")
                sim_registers = sim_registers_cls(memory_id="reg")

            elif o.FAST_REGISTERS in self.options:
                sim_registers_cls = self.plugin_preset.request_plugin("fast_memory")
                sim_registers = sim_registers_cls(memory_id="reg", endness=register_endness)
            else:
                sim_registers_cls = self.plugin_preset.request_plugin("sym_memory")
                sim_registers = sim_registers_cls(memory_id="reg", endness=register_endness)

            # Add registers plugin
            if not self._is_java_jni_project:
                self.register_plugin("registers", sim_registers, inhibit_init=True)

            else:
                # Analog to memory, we add two registers plugins
                native_sim_registers = sim_registers
                javavm_sim_registers_cls = self.plugin_preset.request_plugin("keyvalue_memory")
                javavm_sim_registers = javavm_sim_registers_cls(memory_id="reg")
                self.register_plugin("registers_soot", javavm_sim_registers, inhibit_init=True)
                self.register_plugin("registers_vex", native_sim_registers, inhibit_init=True)

        for p in list(self.plugins.values()):
            p.init_state()

    def __getstate__(self):
        # Don't pickle attributes for plugins. These will be pickled
        # through self._active_plugins.
        s = {k: v for k, v in self.__dict__.items() if k not in self._active_plugins.keys()}
        s["_active_plugins"] = {k: v for k, v in s["_active_plugins"].items() if k not in ("inspect", "regs", "mem")}
        return s

    def __setstate__(self, s):
        self.__dict__.update(s)
        for p in self.plugins.values():
            p.set_state(self)
            if p.STRONGREF_STATE:
                p.set_strongref_state(self)

    def _get_weakref(self):
        return weakref.proxy(self)

    def _get_strongref(self):
        return self

    def __repr__(self):
        try:
            addr = self.addr
            if type(addr) is int:
                ip_str = "%#x" % addr
            else:
                ip_str = repr(addr)
        except (SimValueError, SimSolverModeError):
            ip_str = repr(self.regs.ip)

        return "<SimState @ %s>" % ip_str

    def __setattr__(self, key, value):
        if key == "options":
            # set options
            # this is done to both keep compatibility and make access to .options fast.
            self._set_options(value)
            return
        super().__setattr__(key, value)

    def _set_options(self, v):
        if isinstance(v, (set, list)):
            super().__setattr__("options", SimStateOptions(v))
        elif isinstance(v, SimStateOptions):
            super().__setattr__("options", v)
        else:
            raise SimStateError("Unsupported type '%s' in SimState.options.setter()." % type(v))

    #
    # Easier access to some properties
    #

    @property
    def plugins(self):
        # TODO: This shouldn't be access directly.
        return self._active_plugins

    @property
    def se(self):
        """
        Deprecated alias for `solver`
        """
        global _complained_se
        if not _complained_se:
            _complained_se = True
            l.critical("The name state.se is deprecated; please use state.solver.")
        return self.get_plugin("solver")

    @property
    def ip(self):
        """
        Get the instruction pointer expression, trigger SimInspect breakpoints, and generate SimActions.
        Use ``_ip`` to not trigger breakpoints or generate actions.

        :return: an expression
        """
        return self.regs.ip

    @ip.setter
    def ip(self, val):
        self.regs.ip = val

    @property
    def _ip(self):
        """
        Get the instruction pointer expression without triggering SimInspect breakpoints or generating SimActions.

        :return: an expression
        """
        try:
            return self.regs._ip
        except AttributeError as e:
            raise TypeError(str(e)) from e

    @_ip.setter
    def _ip(self, val):
        """
        Set the instruction pointer without triggering SimInspect breakpoints or generating SimActions.

        :param val: The new instruction pointer.
        :return:    None
        """
        try:
            self.regs._ip = val
        except AttributeError as e:
            raise TypeError(str(e)) from e

    @property
    def addr(self):
        """
        Get the concrete address of the instruction pointer, without triggering SimInspect breakpoints or generating
        SimActions. An integer is returned, or an exception is raised if the instruction pointer is symbolic.

        :return: an int
        """

        ip = self.regs._ip
        if isinstance(ip, SootAddressDescriptor):
            return ip
        return self.solver.eval_one(self.regs._ip)

    @property
    def arch(self) -> Arch:
        if self._is_java_jni_project:
            return self._arch["soot"] if self.ip_is_soot_addr else self._arch["vex"]
        else:
            return self._arch

    @arch.setter
    def arch(self, v):
        self._arch = v

    #
    # Plugin accessors
    #

    def _inspect(self, *args, **kwargs):
        if self.supports_inspect:
            self.inspect.action(*args, **kwargs)

    T = TypeVar("T")

    def _inspect_getattr(self, attr: str, default_value: T):
        if self.supports_inspect:
            if hasattr(self.inspect, attr):
                return getattr(self.inspect, attr)

        return default_value

    #
    # Plugins
    #

    def get_plugin(self, name):
        if self._is_java_jni_project:
            # In case of the JavaVM with JNI support, a state can store the same plugin
            # twice; one for the native and one for the java view of the state.
            suffix = "_soot" if self.ip_is_soot_addr else "_vex"
            name = name + suffix if self.has_plugin(name + suffix) else name
        return super().get_plugin(name)

    def has_plugin(self, name):
        if self._is_java_jni_project:
            # In case of the JavaVM with JNI support, also check for toggled plugins.
            return super().has_plugin(name) or super().has_plugin(name + "_soot")
        return super().has_plugin(name)

    def register_plugin(self, name, plugin, inhibit_init=False):  # pylint: disable=arguments-differ
        # l.debug("Adding plugin %s of type %s", name, plugin.__class__.__name__)
        self._set_plugin_state(plugin, inhibit_init=inhibit_init)
        return super().register_plugin(name, plugin)

    def _init_plugin(self, plugin_cls: Type[SimStatePlugin]) -> SimStatePlugin:
        plugin = plugin_cls()
        self._set_plugin_state(plugin)
        return plugin

    def _set_plugin_state(self, plugin: SimStatePlugin, inhibit_init: bool = False):
        plugin.set_state(self)
        if plugin.STRONGREF_STATE:
            plugin.set_strongref_state(self)
        if not inhibit_init:
            plugin.init_state()

    #
    # Java support
    #

    @property
    def javavm_memory(self):
        """
        In case of an JavaVM with JNI support, a state can store the memory
        plugin twice; one for the native and one for the java view of the state.

        :return: The JavaVM view of the memory plugin.
        """
        if self._is_java_jni_project:
            return self.get_plugin("memory_soot")
        else:
            return self.get_plugin("memory")

    @property
    def javavm_registers(self):
        """
        In case of an JavaVM with JNI support, a state can store the registers
        plugin twice; one for the native and one for the java view of the state.

        :return: The JavaVM view of the registers plugin.
        """
        if self._is_java_jni_project:
            return self.get_plugin("registers_soot")
        else:
            return self.get_plugin("registers")

    #
    # Constraint pass-throughs
    #

    def simplify(self, *args):
        """
        Simplify this state's constraints.
        """
        return self.solver.simplify(*args)

    def add_constraints(self, *args, **kwargs):
        """
        Add some constraints to the state.

        You may pass in any number of symbolic booleans as variadic positional arguments.
        """
        if len(args) > 0 and isinstance(args[0], (list, tuple)):
            raise Exception("Tuple or list passed to add_constraints!")

        if o.TRACK_CONSTRAINTS in self.options and len(args) > 0:
            if o.SIMPLIFY_CONSTRAINTS in self.options:
                constraints = [self.simplify(a) for a in args]
            else:
                constraints = args

            self._inspect("constraints", BP_BEFORE, added_constraints=constraints)
            constraints = self._inspect_getattr("added_constraints", constraints)
            added = self.solver.add(*constraints)
            self._inspect("constraints", BP_AFTER)

            # add actions for the added constraints
            if o.TRACK_CONSTRAINT_ACTIONS in self.options:
                for c in added:
                    sac = SimActionConstraint(self, c)
                    self.history.add_action(sac)
        else:
            # preserve the old action logic for when we don't track constraints (why?)
            if "action" in kwargs and kwargs["action"] and o.TRACK_CONSTRAINT_ACTIONS in self.options and len(args) > 0:
                for arg in args:
                    if self.solver.symbolic(arg):
                        sac = SimActionConstraint(self, arg)
                        self.history.add_action(sac)

        if o.ABSTRACT_SOLVER in self.options and len(args) > 0:
            for arg in args:
                if self.solver.is_false(arg):
                    self._satisfiable = False
                    return

                if self.solver.is_true(arg):
                    continue

                # `is_true` and `is_false` does not use VSABackend currently (see commits 97a75366 and 2dfba73e in
                # claripy). There is a chance that VSA backend can in fact handle it.
                # Therefore we try to resolve it with VSABackend again
                if claripy.backends.vsa.is_false(arg):
                    self._satisfiable = False
                    return

                if claripy.backends.vsa.is_true(arg):
                    continue

                # It's neither True or False. Let's try to apply the condition

                # We take the argument, extract a list of constrained SIs out of it (if we could, of course), and
                # then replace each original SI the intersection of original SI and the constrained one.

                _, converted = self.solver.constraint_to_si(arg)

                for original_expr, constrained_si in converted:
                    if not original_expr.variables:
                        l.error(
                            "Incorrect original_expression to replace in add_constraints(). "
                            "This is due to defects in VSA logics inside claripy. Please report "
                            "to Fish and he will fix it if he's free."
                        )
                        continue

                    new_expr = constrained_si
                    self.registers.replace_all(original_expr, new_expr)
                    self.memory.replace_all(original_expr, new_expr)
                    # tmps
                    temps = self.scratch.temps
                    for idx in range(len(temps)):  # pylint:disable=consider-using-enumerate
                        t = temps[idx]
                        if t is None:
                            continue
                        if t.variables.intersection(original_expr.variables):
                            # replace
                            temps[idx] = t.replace(original_expr, new_expr)

                    l.debug("SimState.add_constraints: Applied to final state.")
        elif o.SYMBOLIC not in self.options and len(args) > 0:
            for arg in args:
                if self.solver.is_false(arg):
                    self._satisfiable = False
                    return

    def satisfiable(self, **kwargs):
        """
        Whether the state's constraints are satisfiable
        """
        if o.ABSTRACT_SOLVER in self.options or o.SYMBOLIC not in self.options:
            extra_constraints = kwargs.pop("extra_constraints", ())
            for e in extra_constraints:
                if self.solver.is_false(e):
                    return False

            return self._satisfiable
        else:
            return self.solver.satisfiable(**kwargs)

    def downsize(self):
        """
        Clean up after the solver engine. Calling this when a state no longer needs to be solved on will reduce memory
        usage.
        """
        if "solver" in self.plugins:
            self.solver.downsize()

    #
    # State branching operations
    #

    def step(self, **kwargs):
        """
        Perform a step of symbolic execution using this state.
        Any arguments to `AngrObjectFactory.successors` can be passed to this.

        :return: A SimSuccessors object categorizing the results of the step.
        """
        return self.project.factory.successors(self, **kwargs)

    def block(self, *args, **kwargs):
        """
        Represent the basic block at this state's instruction pointer.
        Any arguments to `AngrObjectFactory.block` can ba passed to this.

        :return: A Block object describing the basic block of code at this point.
        """
        if not args and "addr" not in kwargs:
            kwargs["addr"] = self.addr
        return self.project.factory.block(*args, backup_state=self, **kwargs)

    # Returns a dict that is a copy of all the state's plugins
    def _copy_plugins(self):
        memo = {}
        out = {}
        for n, p in self._active_plugins.items():
            if id(p) in memo:
                out[n] = memo[id(p)]
            else:
                out[n] = p.copy(memo)
                memo[id(p)] = out[n]

        return out

    def copy(self):
        """
        Returns a copy of the state.
        """

        if self._global_condition is not None:
            raise SimStateError("global condition was not cleared before state.copy().")

        c_plugins = self._copy_plugins()
        state = SimState(
            project=self.project,
            arch=self.arch,
            plugins=c_plugins,
            options=self.options.copy(),
            mode=self.mode,
            os_name=self.os_name,
        )

        if self._is_java_jni_project:
            state.ip_is_soot_addr = self.ip_is_soot_addr

        state.uninitialized_access_handler = self.uninitialized_access_handler
        state._special_memory_filler = self._special_memory_filler
        state.ip_constraints = self.ip_constraints

        return state

    def merge(self, *others, **kwargs):
        """
        Merges this state with the other states. Returns the merging result, merged state, and the merge flag.

        :param states: the states to merge
        :param merge_conditions: a tuple of the conditions under which each state holds
        :param common_ancestor:  a state that represents the common history between the states being merged. Usually it
                                 is only available when EFFICIENT_STATE_MERGING is enabled, otherwise weak-refed states
                                 might be dropped from state history instances.
        :param plugin_whitelist: a list of plugin names that will be merged. If this option is given and is not None,
                                 any plugin that is not inside this list will not be merged, and will be created as a
                                 fresh instance in the new state.
        :param common_ancestor_history:
                                 a SimStateHistory instance that represents the common history between the states being
                                 merged. This is to allow optimal state merging when EFFICIENT_STATE_MERGING is
                                 disabled.
        :return: (merged state, merge flag, a bool indicating if any merging occurred)
        """

        merge_conditions = kwargs.pop("merge_conditions", None)
        common_ancestor = kwargs.pop("common_ancestor", None)
        plugin_whitelist = kwargs.pop("plugin_whitelist", None)
        common_ancestor_history = kwargs.pop("common_ancestor_history", None)

        if len(kwargs) != 0:
            raise ValueError("invalid arguments: %s" % kwargs.keys())

        if merge_conditions is None:
            # TODO: maybe make the length of this smaller? Maybe: math.ceil(math.log(len(others)+1, 2))
            merge_flag = self.solver.BVS("state_merge_%d" % next(merge_counter), 16)
            merge_values = range(len(others) + 1)
            merge_conditions = [merge_flag == b for b in merge_values]
        else:
            merge_conditions = [(self.solver.true if len(mc) == 0 else self.solver.And(*mc)) for mc in merge_conditions]

        if len({o.arch.name for o in others}) != 1:
            raise SimMergeError("Unable to merge due to different architectures.")

        all_plugins = set(self.plugins.keys()) | set.union(*(set(o.plugins.keys()) for o in others))

        if plugin_whitelist is not None:
            all_plugins = all_plugins.intersection(set(plugin_whitelist))

        merged = self.copy()
        merging_occurred = False

        # fix parent
        merged.history.parent = self.history

        # plugins
        for p in all_plugins:
            our_plugin = merged.plugins[p] if p in merged.plugins else None
            their_plugins = [(pl.plugins[p] if p in pl.plugins else None) for pl in others]

            plugin_classes = ({our_plugin.__class__} | {pl.__class__ for pl in their_plugins}) - {None.__class__}
            if len(plugin_classes) != 1:
                raise SimMergeError(f"There are differing plugin classes ({plugin_classes}) for plugin {p}")
            plugin_class = plugin_classes.pop()

            our_filled_plugin = our_plugin if our_plugin is not None else merged.register_plugin(p, plugin_class())
            their_filled_plugins = [
                (tp if tp is not None else t.register_plugin(p, plugin_class())) for t, tp in zip(others, their_plugins)
            ]

            plugin_common_ancestor = (
                common_ancestor.plugins[p] if (common_ancestor is not None and p in common_ancestor.plugins) else None
            )
            if (
                plugin_common_ancestor is None
                and plugin_class is SimStateHistory
                and common_ancestor_history is not None
            ):
                plugin_common_ancestor = common_ancestor_history

            plugin_state_merged = our_filled_plugin.merge(
                their_filled_plugins,
                merge_conditions,
                common_ancestor=plugin_common_ancestor,
            )
            if plugin_state_merged:
                l.debug("Merging occurred in %s", p)
                merging_occurred = True

        merged.add_constraints(merged.solver.Or(*merge_conditions))
        return merged, merge_conditions, merging_occurred

    def widen(self, *others):
        """
        Perform a widening between self and other states
        :param others:
        :return:
        """

        if len({frozenset(o.plugins.keys()) for o in others}) != 1:
            raise SimMergeError("Unable to widen due to different sets of plugins.")
        if len({o.arch.name for o in others}) != 1:
            raise SimMergeError("Unable to widen due to different architectures.")

        widened = self.copy()
        widening_occurred = False

        # plugins
        for p in self.plugins:
            if p in ("solver", "unicorn"):
                continue
            plugin_state_widened = widened.plugins[p].widen([_.plugins[p] for _ in others])
            if plugin_state_widened:
                l.debug("Widening occurred in %s", p)
                widening_occurred = True

        return widened, widening_occurred

    #############################################
    ### Accessors for tmps, registers, memory ###
    #############################################

    def reg_concrete(self, *args, **kwargs):
        """
        Returns the contents of a register but, if that register is symbolic,
        raises a SimValueError.
        """
        e = self.registers.load(*args, **kwargs)
        if self.solver.symbolic(e):
            raise SimValueError("target of reg_concrete is symbolic!")
        return self.solver.eval(e)

    def mem_concrete(self, *args, **kwargs):
        """
        Returns the contents of a memory but, if the contents are symbolic,
        raises a SimValueError.
        """
        e = self.memory.load(*args, **kwargs)
        if self.solver.symbolic(e):
            raise SimValueError("target of mem_concrete is symbolic!")
        return self.solver.eval(e)

    ###############################
    ### Stack operation helpers ###
    ###############################

    @arch_overrideable
    def stack_push(self, thing):
        """
        Push 'thing' to the stack, writing the thing to memory and adjusting the stack pointer.
        """
        # increment sp
        sp = self.regs.sp + self.arch.stack_change
        self.regs.sp = sp
        return self.memory.store(sp, thing, endness=self.arch.memory_endness, size=self.arch.bytes)

    @arch_overrideable
    def stack_pop(self):
        """
        Pops from the stack and returns the popped thing. The length will be the architecture word size.
        """
        sp = self.regs.sp
        self.regs.sp = sp - self.arch.stack_change
        return self.memory.load(sp, self.arch.bytes, endness=self.arch.memory_endness)

    @arch_overrideable
    def stack_read(self, offset, length, bp=False):
        """
        Reads length bytes, at an offset into the stack.

        :param offset:  The offset from the stack pointer.
        :param length:  The number of bytes to read.
        :param bp:      If True, offset from the BP instead of the SP. Default: False.
        """
        sp = self.regs.bp if bp else self.regs.sp
        return self.memory.load(sp + offset, length, endness=self.arch.memory_endness)

    ###############################
    ### Other helpful functions ###
    ###############################

    def make_concrete_int(self, expr):
        if isinstance(expr, int):
            return expr

        if not self.solver.symbolic(expr):
            return self.solver.eval(expr)

        v = self.solver.eval(expr)
        self.add_constraints(expr == v)
        return v

    # This handles the preparation of concrete function launches from abstract functions.
    @arch_overrideable
    def prepare_callsite(self, retval, args, cc="wtf"):
        # TODO
        pass

    def _stack_values_to_string(self, stack_values):
        """
        Convert each stack value to a string

        :param stack_values: A list of values
        :return: The converted string
        """

        strings = []
        for stack_value in stack_values:
            if self.solver.symbolic(stack_value):
                concretized_value = "SYMBOLIC - %s" % repr(stack_value)
            else:
                if len(self.solver.eval_upto(stack_value, 2)) == 2:
                    concretized_value = repr(stack_value)
                else:
                    concretized_value = repr(stack_value)
            strings.append(concretized_value)

        return " .. ".join(strings)

    def dbg_print_stack(self, depth=None, sp=None):
        """
        Only used for debugging purposes.
        Return the current stack info in formatted string. If depth is None, the
        current stack frame (from sp to bp) will be printed out.
        """

        var_size = self.arch.bytes
        sp_sim = self.regs._sp
        bp_sim = self.regs._bp
        if self.solver.symbolic(sp_sim) and sp is None:
            result = "SP is SYMBOLIC"
        elif self.solver.symbolic(bp_sim) and depth is None:
            result = "BP is SYMBOLIC"
        else:
            sp_value = sp if sp is not None else self.solver.eval(sp_sim)
            if self.solver.symbolic(bp_sim):
                result = "SP = 0x%08x, BP is symbolic\n" % (sp_value)
                bp_value = None
            else:
                bp_value = self.solver.eval(bp_sim)
                result = f"SP = 0x{sp_value:08x}, BP = 0x{bp_value:08x}\n"
            if depth is None:
                # bp_value cannot be None here
                depth = (bp_value - sp_value) // var_size + 1  # Print one more value
            pointer_value = sp_value
            for i in range(depth):
                # For AbstractMemory, we wanna utilize more information from VSA
                stack_values = []

                if o.ABSTRACT_MEMORY in self.options:
                    sp = self.regs._sp
                    segment_sizes = self.memory.get_segments(sp + i * var_size, var_size)

                    pos = i * var_size
                    for segment_size in segment_sizes:
                        stack_values.append(self.stack_read(pos, segment_size, bp=False))
                        pos += segment_size
                else:
                    stack_values.append(self.stack_read(i * var_size, var_size, bp=False))

                # Convert it into a big string!
                val = self._stack_values_to_string(stack_values)

                if pointer_value == sp_value:
                    line = f"(sp){pointer_value: 16x} | {val}"
                elif pointer_value == bp_value:
                    line = f"(bp){pointer_value: 16x} | {val}"
                else:
                    line = f"{pointer_value: 20x} | {val}"

                pointer_value += var_size
                result += line + "\n"
        return result

    #
    # Other helper methods
    #

    def set_mode(self, mode):
        self.mode = mode
        self.options = SimStateOptions(o.modes[mode])

    @property
    def thumb(self):
        if not self.arch.name.startswith("ARM"):
            return False

        if self.regs.ip.symbolic:
            # return True when IP can *only* be odd
            new_state = self.copy()
            new_state.add_constraints(new_state.regs.ip % 2 == 1, new_state.regs.ip % 2 != 0)
            return new_state.satisfiable()

        else:
            concrete_ip = self.solver.eval(self.regs.ip)
            return concrete_ip % 2 == 1

    #
    # Some pretty fancy global condition stuff!
    #

    @property
    def with_condition(self):
        @contextlib.contextmanager
        def ctx(c):
            old_condition = self._global_condition
            try:
                new_condition = c if old_condition is None else self.solver.And(old_condition, c)
                self._global_condition = new_condition
                yield
            finally:
                self._global_condition = old_condition

        return ctx

    def _adjust_condition(self, c):
        if self._global_condition is None:
            return c
        elif c is None:
            return self._global_condition
        else:
            return self.solver.And(self._global_condition, c)

    def _adjust_condition_list(self, conditions):
        if self._global_condition is None:
            return conditions
        elif len(conditions) == 0:
            return conditions.__class__((self._global_condition,))
        else:
            return conditions.__class__((self._adjust_condition(self.solver.And(*conditions)),))


default_state_plugin_preset = PluginPreset()
SimState.register_preset("default", default_state_plugin_preset)

from .state_plugins.history import SimStateHistory
from .state_plugins.inspect import BP_AFTER, BP_BEFORE
from .state_plugins.sim_action import SimActionConstraint

from . import sim_options as o
from .errors import SimMergeError, SimValueError, SimStateError, SimSolverModeError

# Type imports for annotations
if TYPE_CHECKING:
    from .storage import DefaultMemory
    from .state_plugins.solver import SimSolver
    from .state_plugins.posix import SimSystemPosix
    from .state_plugins.view import SimRegNameView, SimMemView
    from .state_plugins.callstack import CallStack
    from .state_plugins.inspect import SimInspector
    from .state_plugins.jni_references import SimStateJNIReferences
    from .state_plugins.scratch import SimStateScratch



import logging

l = logging.getLogger(name=__name__)

import itertools

from .plugin import SimStatePlugin


class SimStateLog(SimStatePlugin):
    def __init__(self, log=None):
        SimStatePlugin.__init__(self)

        # general events
        self.events = []

        if log is not None:
            self.events.extend(log.events)

    @property
    def actions(self):
        for e in self.events:
            if isinstance(e, SimAction):
                yield e

    def add_event(self, event_type, **kwargs):
        try:
            new_event = SimEvent(self.state, event_type, **kwargs)
            self.events.append(new_event)
        except TypeError as e:
            raise SimEventError("Exception encountered when logging event") from e

    def _add_event(self, event):
        self.events.append(event)

    def add_action(self, action):
        self.events.append(action)

    def extend_actions(self, new_actions):
        self.events.extend(new_actions)

    def events_of_type(self, event_type):
        return [e for e in self.events if e.type == event_type]

    def actions_of_type(self, action_type):
        return [action for action in self.actions if action.type == action_type]

    @property
    def fresh_constraints(self):
        # pylint: disable=no-member
        return [ev.constraint.ast for ev in self.events if isinstance(ev, SimActionConstraint)]

    @SimStatePlugin.memo
    def copy(self, memo):  # pylint: disable=unused-argument
        return SimStateLog(log=self)

    def _combine(self, others):
        all_events = [e.events for e in itertools.chain([self], others)]
        self.events = [SimEvent(self.state, "merge", event_lists=all_events)]
        return False

    def merge(self, others, merge_conditions, common_ancestor=None):  # pylint: disable=unused-argument
        return self._combine(others)

    def widen(self, others):
        return self._combine(others)

    def clear(self):
        s = self.state
        self.__init__()
        self.state = s
        # self.events = [ ]
        # self.temps.clear()
        # self.used_variables.clear()
        # self.input_variables.clear()


from ..errors import SimEventError
from .sim_event import SimEvent
from .sim_action import SimAction, SimActionConstraint


from angr.sim_state import SimState

SimState.register_default("log", SimStateLog)



import itertools

from ...sim_procedure import SimProcedure
from ...engines.soot.values import SimSootValue_Local, SimSootValue_ParamRef, translate_value
from ...engines.soot.expressions import translate_expr


class JavaSimProcedure(SimProcedure):
    @property
    def is_java(self):
        return True

    @property
    def _engine(self):
        return self.project.factory.default_engine  # FIXME: Currently we assume that it must be a SootMixin

    def _setup_args(self, inst, state, _):
        sim_args = []
        # try to get 'this' reference
        try:
            this_ref = state.javavm_memory.load(addr=SimSootValue_Local("this", None))
            sim_args += [this_ref]
        except KeyError:
            pass
        # fetch all function arguments from memory
        for idx in itertools.count():
            try:
                param = state.javavm_memory.load(addr=SimSootValue_ParamRef(idx, None))
            except KeyError:
                break
            sim_args += [param]
        return sim_args

    def _compute_ret_addr(self, expr):
        return self.state.callstack.ret_addr

    def _prepare_ret_state(self):
        self._engine.prepare_return_state(self.state, self.ret_expr)



import angr
import datetime
import time


class GetSystemTimeAsFileTime(angr.SimProcedure):
    timestamp = None

    def run(self, outptr):
        self.instrument()
        self.state.mem[outptr].qword = self.timestamp

    def instrument(self):
        if angr.options.USE_SYSTEM_TIMES in self.state.options:
            self.fill_from_timestamp(time.time())
        else:
            self.fill_symbolic()

    def fill_from_timestamp(self, ts):
        self.timestamp = int(ts * 1000 * 1000 / 100)
        # convert to microseconds, convert to nanoseconds, convert to 100ns intervals

    def fill_symbolic(self):
        self.timestamp = self.state.solver.BVS("SystemTimeAsFileTime", 64, key=("api", "SystemTimeAsFileTime"))


class GetLocalTime(angr.SimProcedure):
    wYear = None
    wMonth = None
    wDayOfWeek = None
    wDay = None
    wHour = None
    wMinute = None
    wSecond = None
    wMilliseconds = None

    def run(self, outptr):
        self.instrument()
        self.state.mem[outptr].short = self.wYear
        self.state.mem[outptr + 2].short = self.wMonth
        self.state.mem[outptr + 4].short = self.wDayOfWeek
        self.state.mem[outptr + 6].short = self.wDay
        self.state.mem[outptr + 8].short = self.wHour
        self.state.mem[outptr + 10].short = self.wMinute
        self.state.mem[outptr + 12].short = self.wSecond
        self.state.mem[outptr + 14].short = self.wMilliseconds

    def instrument(self):
        """
        Override this method, setting the class vars of the same names as the windows SYSTEMTIME structure.
        By default, fills it with appropriately constrained symbolic data, but makes no attempt to make sure that the
        value can make sense semantically, e.g. feb 31 can be represented.

        ...unless the USE_SYSTEM_TIMES state option is present, in which case it will use the system time.

        https://msdn.microsoft.com/en-us/library/windows/desktop/ms724950(v=vs.85).aspx
        """
        if angr.options.USE_SYSTEM_TIMES in self.state.options:
            self.fill_from_timestamp(time.time())
        else:
            self.fill_symbolic()

    def fill_symbolic(self):
        """
        Fill the class with constrained symbolic values.
        """
        self.wYear = self.state.solver.BVS("cur_year", 16, key=("api", "GetLocalTime", "cur_year"))
        self.wMonth = self.state.solver.BVS("cur_month", 16, key=("api", "GetLocalTime", "cur_month"))
        self.wDayOfWeek = self.state.solver.BVS("cur_dayofweek", 16, key=("api", "GetLocalTime", "cur_dayofweek"))
        self.wDay = self.state.solver.BVS("cur_day", 16, key=("api", "GetLocalTime", "cur_day"))
        self.wHour = self.state.solver.BVS("cur_hour", 16, key=("api", "GetLocalTime", "cur_hour"))
        self.wMinute = self.state.solver.BVS("cur_minute", 16, key=("api", "GetLocalTime", "cur_minute"))
        self.wSecond = self.state.solver.BVS("cur_second", 16, key=("api", "GetLocalTime", "cur_second"))
        self.wMilliseconds = self.state.solver.BVS(
            "cur_millisecond", 16, key=("api", "GetLocalTime", "cur_millisecond")
        )

        self.state.add_constraints(self.wYear >= 1601)
        self.state.add_constraints(self.wYear <= 30827)
        self.state.add_constraints(self.wMonth >= 1)
        self.state.add_constraints(self.wMonth <= 12)
        self.state.add_constraints(self.wDayOfWeek <= 6)
        self.state.add_constraints(self.wDay >= 1)
        self.state.add_constraints(self.wDay <= 31)
        self.state.add_constraints(self.wHour <= 23)
        self.state.add_constraints(self.wMinute <= 59)
        self.state.add_constraints(self.wSecond <= 59)
        self.state.add_constraints(self.wMilliseconds <= 999)

    def fill_from_timestamp(self, ts):
        """
        Fill the class with the appropriate values extracted from the given timestamp.

        :param ts:  A POSIX timestamp.
        """
        dt = datetime.datetime.fromtimestamp(ts)
        self.wYear = dt.year
        self.wMonth = dt.month
        self.wDayOfWeek = dt.isoweekday() % 7  # :/
        self.wDay = dt.day
        self.wHour = dt.hour
        self.wMinute = dt.minute
        self.wSecond = dt.second
        self.wMilliseconds = dt.microsecond // 1000


class QueryPerformanceCounter(angr.SimProcedure):
    def run(self, ptr):
        if angr.options.USE_SYSTEM_TIMES in self.state.options:
            val = int(time.clock() * 1000000) + 12345678
            self.state.mem[ptr].qword = val
        else:
            self.state.mem[ptr].qword = self.state.solver.BVS(
                "QueryPerformanceCounter_result", 64, key=("api", "QueryPerformanceCounter")
            )
        return 1


class GetTickCount(angr.SimProcedure):
    def run(self):
        if angr.options.USE_SYSTEM_TIMES in self.state.options:
            return int(time.clock() * 1000) + 12345
        else:
            val = self.state.solver.BVS("GetTickCount_result", 32, key=("api", "GetTickCount"))
            return val


class GetTickCount64(angr.SimProcedure):
    KEY = ("sim_time", "GetTickCount")

    def run(self):
        if angr.options.USE_SYSTEM_TIMES in self.state.options:
            return int(time.clock() * 1000) + 12345
        else:
            return self.state.solver.BVS("GetTickCount64_result", 64, key=("api", "GetTickCount64"))



# pylint:disable=line-too-long
import logging

from ...sim_type import SimTypeFunction,     SimTypeShort, SimTypeInt, SimTypeLong, SimTypeLongLong, SimTypeDouble, SimTypeFloat,     SimTypePointer,     SimTypeChar,     SimStruct,     SimTypeFixedSizeArray,     SimTypeBottom,     SimUnion,     SimTypeBool
from ...calling_conventions import SimCCStdcall, SimCCMicrosoftAMD64
from .. import SIM_PROCEDURES as P
from . import SimLibrary


_l = logging.getLogger(name=__name__)


lib = SimLibrary()
lib.set_default_cc('X86', SimCCStdcall)
lib.set_default_cc('AMD64', SimCCMicrosoftAMD64)
lib.set_library_names("rasdlg.dll")
prototypes = \
    {
        #
        'RasPhonebookDlgA': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimStruct({"dwSize": SimTypeInt(signed=False, label="UInt32"), "hwndOwner": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "dwFlags": SimTypeInt(signed=False, label="UInt32"), "xDlg": SimTypeInt(signed=True, label="Int32"), "yDlg": SimTypeInt(signed=True, label="Int32"), "dwCallbackId": SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), "pCallback": SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeBottom(label="Void"), arg_names=["param0", "param1", "param2", "param3"]), offset=0), "dwError": SimTypeInt(signed=False, label="UInt32"), "reserved": SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), "reserved2": SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0)}, name="RASPBDLGA", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["lpszPhonebook", "lpszEntry", "lpInfo"]),
        #
        'RasPhonebookDlgW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimStruct({"dwSize": SimTypeInt(signed=False, label="UInt32"), "hwndOwner": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "dwFlags": SimTypeInt(signed=False, label="UInt32"), "xDlg": SimTypeInt(signed=True, label="Int32"), "yDlg": SimTypeInt(signed=True, label="Int32"), "dwCallbackId": SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), "pCallback": SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeBottom(label="Void"), arg_names=["param0", "param1", "param2", "param3"]), offset=0), "dwError": SimTypeInt(signed=False, label="UInt32"), "reserved": SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), "reserved2": SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0)}, name="RASPBDLGW", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["lpszPhonebook", "lpszEntry", "lpInfo"]),
        #
        'RasEntryDlgA': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimStruct({"dwSize": SimTypeInt(signed=False, label="UInt32"), "hwndOwner": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "dwFlags": SimTypeInt(signed=False, label="UInt32"), "xDlg": SimTypeInt(signed=True, label="Int32"), "yDlg": SimTypeInt(signed=True, label="Int32"), "szEntry": SimTypeFixedSizeArray(SimTypeBottom(label="CHAR"), 257), "dwError": SimTypeInt(signed=False, label="UInt32"), "reserved": SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), "reserved2": SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0)}, name="RASENTRYDLGA", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["lpszPhonebook", "lpszEntry", "lpInfo"]),
        #
        'RasEntryDlgW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimStruct({"dwSize": SimTypeInt(signed=False, label="UInt32"), "hwndOwner": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "dwFlags": SimTypeInt(signed=False, label="UInt32"), "xDlg": SimTypeInt(signed=True, label="Int32"), "yDlg": SimTypeInt(signed=True, label="Int32"), "szEntry": SimTypeFixedSizeArray(SimTypeChar(label="Char"), 257), "dwError": SimTypeInt(signed=False, label="UInt32"), "reserved": SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), "reserved2": SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0)}, name="RASENTRYDLGW", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["lpszPhonebook", "lpszEntry", "lpInfo"]),
        #
        'RasDialDlgA': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimStruct({"dwSize": SimTypeInt(signed=False, label="UInt32"), "hwndOwner": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "dwFlags": SimTypeInt(signed=False, label="UInt32"), "xDlg": SimTypeInt(signed=True, label="Int32"), "yDlg": SimTypeInt(signed=True, label="Int32"), "dwSubEntry": SimTypeInt(signed=False, label="UInt32"), "dwError": SimTypeInt(signed=False, label="UInt32"), "reserved": SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), "reserved2": SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0)}, name="RASDIALDLG", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["lpszPhonebook", "lpszEntry", "lpszPhoneNumber", "lpInfo"]),
        #
        'RasDialDlgW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimStruct({"dwSize": SimTypeInt(signed=False, label="UInt32"), "hwndOwner": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "dwFlags": SimTypeInt(signed=False, label="UInt32"), "xDlg": SimTypeInt(signed=True, label="Int32"), "yDlg": SimTypeInt(signed=True, label="Int32"), "dwSubEntry": SimTypeInt(signed=False, label="UInt32"), "dwError": SimTypeInt(signed=False, label="UInt32"), "reserved": SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), "reserved2": SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0)}, name="RASDIALDLG", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["lpszPhonebook", "lpszEntry", "lpszPhoneNumber", "lpInfo"]),
    }

lib.set_prototypes(prototypes)



# pylint:disable=no-self-use,arguments-renamed
import enum
from typing import DefaultDict, Any, List, Union, Dict, Tuple, Set, Optional
from collections import OrderedDict, defaultdict

import ailment

from ....utils.constants import SWITCH_MISSING_DEFAULT_NODE_ADDR
from ..structuring.structurer_nodes import SwitchCaseNode, ConditionNode, SequenceNode, MultiNode, BaseNode, BreakNode
from ..sequence_walker import SequenceWalker
from ..condition_processor import ConditionProcessor, EmptyBlockNotice
from ..utils import is_statement_terminating


class CmpOp(enum.Enum):
    """
    All supported comparison operators.
    """

    LT = 0
    GT = 1
    EQ = 2
    NE = 3


class ConditionalRegion:
    """
    Describes a conditional region.
    """

    __slots__ = (
        "variable",
        "op",
        "value",
        "node",
        "parent",
    )

    def __init__(self, variable, op: CmpOp, value: int, node: Union[ConditionNode, ailment.Block], parent=None):
        self.variable = variable
        self.op = op
        self.value = value
        self.node = node
        self.parent = parent

    def __repr__(self):
        return f"<{self.variable} {self.op} {self.value} @ {self.node.addr:#x}, parent node {self.parent}>"


class SwitchCaseRegion:
    """
    Describes an already-recovered switch region.
    """

    __slots__ = (
        "variable",
        "node",
        "parent",
    )

    def __init__(self, variable, node: SwitchCaseNode, parent=None):
        self.variable = variable
        self.node = node
        self.parent = parent

    def __repr__(self):
        return f"<{self.variable} ({len(self.node.cases)} cases) @ {self.node.addr:#x}, parent node {self.parent}>"


class SwitchClusterFinder(SequenceWalker):
    """
    Find comparisons and switches in order to identify switch clusters.
    """

    def __init__(self, node):
        handlers = {
            SwitchCaseNode: self._handle_SwitchCase,
            ConditionNode: self._handle_Condition,
            ailment.Block: self._handle_Block,
        }
        super().__init__(handlers)

        self.var2condnodes: DefaultDict[Any, List[ConditionalRegion]] = defaultdict(list)
        self.var2switches: DefaultDict[Any, List[SwitchCaseRegion]] = defaultdict(list)

        self.walk(node)

    def _handle_Block(self, node: ailment.Block, parent=None, **kwargs):  # pylint:disable=unused-argument
        if node.statements and isinstance(node.statements[-1], ailment.Stmt.ConditionalJump):
            cond = node.statements[-1].condition
            self._process_condition(cond, node, parent)

    def _handle_Condition(self, node, parent=None, **kwargs):
        cond = node.condition
        self._process_condition(cond, node, parent)
        return super()._handle_Condition(node, parent=parent, **kwargs)

    def _handle_SwitchCase(self, node: SwitchCaseNode, parent=None, **kwargs):
        cond = node.switch_expr
        if hasattr(cond, "variable"):
            variable = cond.variable
            scr = SwitchCaseRegion(variable, node, parent)
            self.var2switches[variable].append(scr)
        return super()._handle_SwitchCase(node, parent=parent, **kwargs)

    def _process_condition(self, cond: ailment.Expr.Expression, node: Union[ConditionNode, ailment.Block], parent):
        negated = False
        if isinstance(cond, ailment.Expr.UnaryOp) and cond.op == "Not":
            negated = True
            cond = cond.operand
        if isinstance(cond, ailment.Expr.BinaryOp) and cond.op in {
            "CmpLE",
            "CmpLT",
            "CmpGE",
            "CmpGT",
            "CmpEQ",
            "CmpNE",
        }:
            v = None
            variable = None
            if isinstance(cond.operands[1], ailment.Expr.Const):
                v = cond.operands[1].value
            if isinstance(cond.operands[0], (ailment.Expr.Register, ailment.Expr.Load)):
                if hasattr(cond.operands[0], "variable"):
                    # there we go
                    variable = cond.operands[0].variable

            if v is not None and variable is not None:
                if negated:
                    real_op = ailment.Expr.BinaryOp.COMPARISON_NEGATION[cond.op]
                else:
                    real_op = cond.op
                # eliminate equal
                if real_op == "CmpLE":
                    real_op = CmpOp.LT
                    v += 1
                elif real_op == "CmpGE":
                    real_op = CmpOp.GT
                    v -= 1
                elif real_op == "CmpLT":
                    real_op = CmpOp.LT
                elif real_op == "CmpGT":
                    real_op = CmpOp.GT
                elif real_op == "CmpEQ":
                    real_op = CmpOp.EQ
                elif real_op == "CmpNE":
                    real_op = CmpOp.NE
                else:
                    raise TypeError(f"Unsupported real_op value {real_op}")

                cr = ConditionalRegion(variable, real_op, v, node, parent=parent)
                self.var2condnodes[variable].append(cr)


class SwitchClusterReplacer(SequenceWalker):
    """
    Replace an identified switch cluster with a newly created SwitchCase node.
    """

    def __init__(self, region, to_replace, replace_with):
        handlers = {
            ConditionNode: self._handle_Condition,
        }
        super().__init__(handlers)

        self.to_replace = to_replace
        self.replace_with = replace_with
        self.walk(region)

    def _handle_Condition(self, node, **kwargs):
        if node is self.to_replace:
            return self.replace_with
        return super()._handle_Condition(node, **kwargs)


def is_simple_jump_node(node, case_addrs, targets: Optional[Set[int]] = None) -> bool:
    if isinstance(node, (SequenceNode, MultiNode)):
        return all(is_simple_jump_node(nn, case_addrs) for nn in node.nodes)
    if isinstance(node, ailment.Block):
        for stmt in node.statements:
            if isinstance(stmt, ailment.Stmt.Label):
                pass
            elif (
                isinstance(stmt, ailment.Stmt.Jump)
                and isinstance(stmt.target, ailment.Expr.Const)
                and stmt.target.value in case_addrs
            ):
                if targets is not None:
                    targets.add(stmt.target.value)
            elif isinstance(stmt, ailment.Stmt.ConditionalJump):
                ok = False
                if (
                    stmt.true_target is None
                    or isinstance(stmt.true_target, ailment.Expr.Const)
                    and stmt.true_target.value in case_addrs
                ):
                    ok = True
                    if stmt.true_target is not None and targets is not None:
                        targets.add(stmt.true_target.value)
                if (
                    stmt.false_target is None
                    or isinstance(stmt.false_target, ailment.Expr.Const)
                    and stmt.false_target.value in case_addrs
                ):
                    ok = True
                    if stmt.false_target is not None and targets is not None:
                        targets.add(stmt.false_target.value)
                if not ok:
                    return False
            else:
                return False
        return True
    return False


def filter_cond_regions(cond_regions: List[ConditionalRegion], case_addrs: Set[int]) -> List[ConditionalRegion]:
    """
    Remove all conditional regions that cannot be merged into switch(es).
    """

    # TODO: the check is too strict!

    new_cond_regions = []
    for r in cond_regions:
        ok = True
        if isinstance(r.node, ConditionNode):
            true_node, false_node = r.node.true_node, r.node.false_node
            # it should only contain a jump into one of the cases
            ok = (true_node is None or is_simple_jump_node(true_node, case_addrs)) and (
                false_node is None or is_simple_jump_node(false_node, case_addrs)
            )
        elif isinstance(r.node, ailment.Block):
            if not is_simple_jump_node(r.node, case_addrs):
                ok = False
        else:
            raise TypeError(f"Unsupported r.node type {type(r.node)}")

        if ok:
            new_cond_regions.append(r)

    return new_cond_regions


def update_switch_case_list(
    cases: List[Tuple[Union[int, Tuple[int, ...]], SequenceNode]],
    old_case_id: Union[int, Tuple[int, ...]],
    new_case_id: int,
) -> None:
    """
    Update cases in-place. Make new_case_id directly jump to old_case_id.
    """

    for i, (case_ids, case_node) in list(enumerate(cases)):
        if isinstance(case_ids, int):
            match = old_case_id == case_ids
            new_key = (case_ids, new_case_id)
        elif isinstance(case_ids, tuple):
            match = old_case_id == case_ids or old_case_id in case_ids
            new_key = case_ids + (new_case_id,)
        else:
            raise TypeError(f"Unsupported case_ids type {type(case_ids)}")
        if match:
            cases[i] = (new_key, case_node)
            break


def simplify_switch_clusters(
    region, var2condnodes: Dict[Any, List[ConditionalRegion]], var2switches: Dict[Any, List[SwitchCaseRegion]]
):
    """
    Identify switch clusters and simplify each of them.

    :param region:          The region to simplify.
    :param var2condnodes:   A dict that stores the mapping from (potential) switch variables to conditional regions.
    :param var2switches:    A dict that stores the mapping from switch variables to switch-case regions.
    :return:                None
    """

    for variable in var2switches:
        switch_regions = var2switches[variable]
        cond_regions = list(var2condnodes[variable])

        if not cond_regions:
            continue

        # each switch region belongs to a conditional region
        switch_region_to_parent_region: Dict[SwitchCaseRegion, Tuple[ConditionalRegion, str]] = {}
        used_condnodes_and_branch = set()
        for r in switch_regions:
            for cr in cond_regions:
                if isinstance(cr.node, ConditionNode):
                    if cr.node.true_node is r.node:
                        switch_region_to_parent_region[r] = (cr, "true")
                        used_condnodes_and_branch.add((cr, "true"))
                        break
                    if cr.node.false_node is r.node:
                        switch_region_to_parent_region[r] = (cr, "false")
                        used_condnodes_and_branch.add((cr, "false"))
                        break
        if len(switch_region_to_parent_region) != len(switch_regions):
            continue

        # there is at most one default node
        default_node_addrs = set()
        for r in switch_regions:
            if r.node.default_node is not None:
                dn = r.node.default_node
                if dn.addr == SWITCH_MISSING_DEFAULT_NODE_ADDR:
                    # parse the jump target out of the manually created default node
                    if (
                        isinstance(dn, ailment.Block)
                        and isinstance(dn.statements[-1], ailment.Stmt.Jump)
                        and isinstance(dn.statements[-1].target, ailment.Expr.Const)
                    ):
                        default_node_addrs.add(dn.statements[-1].target.value)
                    else:
                        default_node_addrs.add(dn.addr)
                else:
                    default_node_addrs.add(dn.addr)
        if len(default_node_addrs) > 1:
            continue

        # ensure cases in each switch do not overlap
        case_ids = set()
        overlaps = False
        for r in switch_regions:
            for case_id in r.node.cases:
                if case_id in case_ids:
                    overlaps = True
                    break
                case_ids.add(case_id)
            if overlaps:
                break
        if overlaps:
            continue

        used_condnodes = list(condnode for condnode, _ in used_condnodes_and_branch)

        # collect addresses for all case nodes
        case_addr_to_case_id: Dict[int, Union[int, Tuple[int, ...], str]] = {}
        for sr in switch_regions:
            for case_id, case_node in sr.node.cases.items():
                case_addr_to_case_id[case_node.addr] = case_id
            if sr.node.default_node is not None:
                case_addr_to_case_id[sr.node.default_node.addr] = "default"
        case_addrs: Set[int] = set(case_addr_to_case_id)

        # filter cond_regions
        mergeable_cond_regions = filter_cond_regions(cond_regions, case_addrs)

        # list all unmatched conditional nodes
        standalone_condnodes: List[Tuple[ConditionalRegion, str]] = []
        standalone_condjumps: List[ConditionalRegion] = []
        for cr in mergeable_cond_regions:
            if isinstance(cr.node, ConditionNode):
                if cr.node.true_node is not None and (cr, "true") not in used_condnodes_and_branch:
                    standalone_condnodes.append((cr, "true"))
                if cr.node.false_node is not None and (cr, "false") not in used_condnodes_and_branch:
                    standalone_condnodes.append((cr, "false"))
            elif isinstance(cr.node, ailment.Block):
                # the last statement is a conditional jump
                standalone_condjumps.append(cr)
            else:
                raise TypeError(f"Unsupported cr.node type {type(cr.node)}")

        # all conditional regions belong to the same parent node
        parents = {r.parent for r in mergeable_cond_regions + used_condnodes}
        if len(parents) != 1:
            continue

        # the hard part: ensure there are no gaps between value ranges of each conditional region

        parent = next(iter(parents))

        # build the switch-case node
        cases = []
        default_node = None

        # first let's merge all switch cases
        # to ensure proper handling of fallthrough into the default node, we add the one with a default node at the end
        switch_regions_no_default_nodes = []
        switch_regions_default_nodes = []
        for sr in switch_regions:
            if sr.node.default_node is not None and not switch_regions_default_nodes:
                switch_regions_default_nodes.append(sr)
            else:
                switch_regions_no_default_nodes.append(sr)

        for sr in switch_regions_no_default_nodes:
            for case_idx, node in sr.node.cases.items():
                cases.append((case_idx, node))
        if switch_regions_default_nodes:
            for case_idx, node in switch_regions_default_nodes[0].node.cases.items():
                cases.append((case_idx, node))
            default_node = switch_regions_default_nodes[0].node.default_node

        # then let's handle conditional nodes and jumps

        # for now, we only support two types:
        #
        # if (var == A) {
        #     goto one_case/default;
        # }
        # ... the actual switch
        #
        #  or
        #
        # if (var > A) {
        #    goto default;
        # }
        # ... the actual switch

        ok = False
        if not standalone_condnodes:
            if not standalone_condjumps:
                ok = True
            elif len(standalone_condjumps) == 1:
                condjump = standalone_condjumps[0]
                targets = set()
                if is_simple_jump_node(condjump.node, case_addrs, targets=targets) and len(targets) == 1:
                    target = next(iter(targets))
                    if condjump.op == CmpOp.EQ:
                        # which case is it jumping to?
                        existing_case_id = case_addr_to_case_id[target]
                        if existing_case_id == "default":
                            # just add it at the end
                            cases.append((condjump.value, SequenceNode(0, nodes=[])))
                        else:
                            update_switch_case_list(
                                cases,
                                existing_case_id,
                                condjump.value,
                            )
                        ok = True
                    elif condjump.op in {CmpOp.GT, CmpOp.LT}:
                        if default_node is not None and target == default_node.addr:
                            # it goes to default anyway - we don't need to do anything
                            ok = True

        if not ok:
            continue

        # build the SwitchCase node and replace old nodes in the parent node
        cases_dict = OrderedDict(cases)
        new_switchcase = SwitchCaseNode(
            switch_regions_default_nodes[0].node.switch_expr,
            cases_dict,
            default_node,
            addr=switch_regions_default_nodes[0].node.addr,
        )

        # what are we trying to replace?
        if isinstance(parent, SequenceNode):
            start_idx = None  # inclusive
            end_idx = None  # not inclusive
            for idx, node in enumerate(parent.nodes):
                if (
                    any(cr.node is node for cr, _ in standalone_condnodes)
                    or any(cr.node is node for cr in standalone_condjumps)
                    or any(cr.node is node for cr in used_condnodes)
                ):
                    if start_idx is None:
                        start_idx = idx
                    if end_idx is not None:
                        # found a gap...
                        pass
                else:
                    if start_idx is not None:
                        if end_idx is None:
                            end_idx = idx

            if start_idx is None:
                continue
            if end_idx is None:
                end_idx = len(parent.nodes)

            # we want to replace parent.nodes[start_idx : end_idx] with a SwitchCaseNode

            parent.nodes[start_idx:end_idx] = [new_switchcase]

        elif isinstance(parent, ConditionNode):
            # replace this condition node with the new SwitchCase node
            SwitchClusterReplacer(region, parent, new_switchcase)

        else:
            # unsupported for now
            continue


def simplify_lowered_switches(region: SequenceNode, var2condnodes: Dict[Any, List[ConditionalRegion]], functions):
    """
    Identify a lowered switch and simplify it into a switch-case if possible.

    :param region:          The region to simplify.
    :param var2condnodes:   A dict that stores the mapping from (potential) switch variables to conditional regions.
    :return:                None
    """

    for var, condnodes in var2condnodes.items():
        # an arbitrary threshold of 8, and must have at least one > or <
        if len(condnodes) > 8 and any(condnode.op in {CmpOp.GT, CmpOp.LT} for condnode in condnodes):
            simplify_lowered_switches_core(region, var, condnodes, functions)


def simplify_lowered_switches_core(
    region: SequenceNode, var, condnodes, functions  # pylint:disable=unused-argument
) -> bool:
    node_to_condnode = {}
    parent_node_to_condnodes = defaultdict(list)

    for condnode in condnodes:
        if condnode.node in node_to_condnode:
            return False
        node_to_condnode[condnode.node] = condnode
        parent_node_to_condnodes[condnode.parent].append(condnode)

    first_node_finder = FindFirstNodeInSet(set(node_to_condnode))
    first_node_finder.walk(region)
    outermost_node = first_node_finder.first_node

    if outermost_node is None:
        return False

    caseno_to_node = {}
    default_node_candidates: List[Tuple[BaseNode, BaseNode]] = []  # parent to default node candidate
    stack: List[(ConditionNode, int, int)] = [(outermost_node, 0, 0xFFFF_FFFF_FFFF_FFFF)]
    while stack:
        node, min_, max_ = stack.pop(0)
        if node not in node_to_condnode:
            return False
        if not isinstance(node, ConditionNode):
            # not fully structured
            return False
        condnode = node_to_condnode[node]
        if condnode.op == CmpOp.EQ:
            if node.true_node is not None:
                # true node is the case
                if condnode.value in caseno_to_node:
                    # overlapping case
                    return False
                caseno_to_node[condnode.value] = node.true_node

            if node.false_node is not None:
                # is it the default node?
                if node.false_node not in node_to_condnode:
                    default_node_candidates.append((node, node.false_node))
                else:
                    stack.append((node.false_node, min_, max_))
        elif condnode.op == CmpOp.GT:
            # TODO: Handle cases with multiple case numbers
            if node.true_node is not None:
                stack.append((node.true_node, condnode.value + 1, max_))
            if node.false_node is not None:
                stack.append((node.false_node, min_, condnode.value))
        elif condnode.op == CmpOp.LT:
            # TODO: Handle cases with multiple case numbers
            if node.true_node is not None:
                stack.append((node.true_node, min_, condnode.value - 1))
            if node.false_node is not None:
                stack.append((node.false_node, condnode.value, max_))
        elif condnode.op == CmpOp.NE:
            if node.false_node is not None:
                # false node is the case
                if condnode.value in caseno_to_node:
                    # overlapping case
                    return False
                caseno_to_node[condnode.value] = node.false_node

            if node.true_node is not None:
                # is it the default node?
                if node.true_node not in node_to_condnode:
                    default_node_candidates.append((node, node.true_node))
                else:
                    stack.append((node.true_node, min_, max_))

    # sanity check: all default node candidates should have the same address
    # note that this sanity check is not strict. we sincerely hope that prior optimizations haven't made these default
    # node duplicates different from each other
    # TODO: Implement a stricter check
    if len({candidate.addr for _, candidate in default_node_candidates}) > 1:
        return False
    default_node = None
    if default_node_candidates:
        default_node = default_node_candidates[0][1]

    # build the switch-case
    cases = OrderedDict()
    for k in sorted(caseno_to_node):
        case_node = caseno_to_node[k]
        try:
            last_stmts = ConditionProcessor.get_last_statements(case_node)
        except EmptyBlockNotice:
            # unexpected
            return False
        if not all(is_statement_terminating(stmt, functions) for stmt in last_stmts):
            # insert a break node at the end
            case_node = SequenceNode(
                case_node.addr,
                nodes=[
                    case_node,
                    BreakNode(case_node.addr, None),
                ],
            )
        cases[k] = case_node

    if not isinstance(outermost_node.condition.operands[0], ailment.Expr.Const) and isinstance(
        outermost_node.condition.operands[1], ailment.Expr.Const
    ):
        switch_expr = outermost_node.condition.operands[0]
    elif not isinstance(outermost_node.condition.operands[1], ailment.Expr.Const) and isinstance(
        outermost_node.condition.operands[0], ailment.Expr.Const
    ):
        switch_expr = outermost_node.condition.operands[1]
    else:
        return False

    new_switchcase = SwitchCaseNode(
        switch_expr,
        cases,
        default_node,
        addr=outermost_node.addr,
    )

    # replace this condition node with the new SwitchCase node
    SwitchClusterReplacer(region, outermost_node, new_switchcase)

    return True


class FindFirstNodeInSet(SequenceWalker):
    """
    Find the first node out of a set of node appearing in a SequenceNode (and its tree).
    """

    def __init__(self, node_set: Set[BaseNode]):
        super().__init__(update_seqnode_in_place=False, force_forward_scan=True)
        self.node_set = node_set
        self.first_node = None

    def _handle(self, node, **kwargs):
        if node in self.node_set:
            self.first_node = node
        if self.first_node is None:
            super()._handle(node, **kwargs)



import logging

l = logging.getLogger(name=__name__)

from .plugin import SimStatePlugin
from ..errors import SimUCManagerAllocationError


class SimUCManager(SimStatePlugin):
    def __init__(self, man=None):
        SimStatePlugin.__init__(self)

        if man:
            self._region_base = man._region_base
            self._pos = man._pos
            self._alloc_depth_map = man._alloc_depth_map.copy()

        else:
            self._region_base = None  # It will be set later when self.state is set
            self._pos = 0

            self._alloc_depth_map = {}

        #
        # Some constants
        #

        # The size of each region, in bytes
        self._region_size = 0x1000
        # The maximum allocation depth
        self._max_alloc_depth = 20

    def assign(self, dst_addr_ast):
        """
        Assign a new region for under-constrained symbolic execution.

        :param dst_addr_ast: the symbolic AST which address of the new allocated region will be assigned to.
        :return: as ast of memory address that points to a new region
        """

        if dst_addr_ast.uc_alloc_depth > self._max_alloc_depth:
            raise SimUCManagerAllocationError(
                "Current allocation depth %d is greater than the cap (%d)"
                % (dst_addr_ast.uc_alloc_depth, self._max_alloc_depth)
            )

        abs_addr = self._region_base + self._pos
        ptr = self.state.solver.BVV(abs_addr, self.state.arch.bits)
        self._pos += self._region_size

        self._alloc_depth_map[(abs_addr - self._region_base) // self._region_size] = dst_addr_ast.uc_alloc_depth

        l.debug("Assigned new memory region %s", ptr)
        return ptr

    @SimStatePlugin.memo
    def copy(self, memo):  # pylint: disable=unused-argument
        return SimUCManager(man=self)

    def get_alloc_depth(self, addr):
        block_pos = (addr - self._region_base) // self._region_size

        if block_pos not in self._alloc_depth_map:
            return None

        return self._alloc_depth_map[block_pos]

    def is_bounded(self, ast):
        """
        Test whether an AST is bounded by any existing constraint in the related solver.

        :param ast: an claripy.AST object
        :return: True if there is at least one related constraint, False otherwise
        """

        return len(ast.variables.intersection(self.state.solver._solver.variables)) != 0

    def set_state(self, state):
        super().set_state(state)
        self._region_base = 0xD0 << (self.state.arch.bits - 8)


from angr.sim_state import SimState

SimState.register_default("uc_manager", SimUCManager)



# pylint:disable=line-too-long
import logging

from ...sim_type import SimTypeFunction,     SimTypeShort, SimTypeInt, SimTypeLong, SimTypeLongLong, SimTypeDouble, SimTypeFloat,     SimTypePointer,     SimTypeChar,     SimStruct,     SimTypeFixedSizeArray,     SimTypeBottom,     SimUnion,     SimTypeBool
from ...calling_conventions import SimCCStdcall, SimCCMicrosoftAMD64
from .. import SIM_PROCEDURES as P
from . import SimLibrary


_l = logging.getLogger(name=__name__)


lib = SimLibrary()
lib.set_default_cc('X86', SimCCStdcall)
lib.set_default_cc('AMD64', SimCCMicrosoftAMD64)
lib.set_library_names("mswsock.dll")
prototypes = \
    {
        #
        'WSARecvEx': SimTypeFunction([SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["s", "buf", "len", "flags"]),
        #
        'TransmitFile': SimTypeFunction([SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"Internal": SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), "InternalHigh": SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), "Anonymous": SimUnion({"Anonymous": SimStruct({"Offset": SimTypeInt(signed=False, label="UInt32"), "OffsetHigh": SimTypeInt(signed=False, label="UInt32")}, name="_Anonymous_e__Struct", pack=False, align=None), "Pointer": SimTypePointer(SimTypeBottom(label="Void"), offset=0)}, name="<anon>", label="None"), "hEvent": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)}, name="OVERLAPPED", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"Head": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "HeadLength": SimTypeInt(signed=False, label="UInt32"), "Tail": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "TailLength": SimTypeInt(signed=False, label="UInt32")}, name="TRANSMIT_FILE_BUFFERS", pack=False, align=None), offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["hSocket", "hFile", "nNumberOfBytesToWrite", "nNumberOfBytesPerSend", "lpOverlapped", "lpTransmitBuffers", "dwReserved"]),
        #
        'AcceptEx': SimTypeFunction([SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimStruct({"Internal": SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), "InternalHigh": SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), "Anonymous": SimUnion({"Anonymous": SimStruct({"Offset": SimTypeInt(signed=False, label="UInt32"), "OffsetHigh": SimTypeInt(signed=False, label="UInt32")}, name="_Anonymous_e__Struct", pack=False, align=None), "Pointer": SimTypePointer(SimTypeBottom(label="Void"), offset=0)}, name="<anon>", label="None"), "hEvent": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)}, name="OVERLAPPED", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["sListenSocket", "sAcceptSocket", "lpOutputBuffer", "dwReceiveDataLength", "dwLocalAddressLength", "dwRemoteAddressLength", "lpdwBytesReceived", "lpOverlapped"]),
        #
        'GetAcceptExSockaddrs': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimStruct({"sa_family": SimTypeShort(signed=False, label="UInt16"), "sa_data": SimTypeFixedSizeArray(SimTypeBottom(label="CHAR"), 14)}, name="SOCKADDR", pack=False, align=None), offset=0), offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0), SimTypePointer(SimTypePointer(SimStruct({"sa_family": SimTypeShort(signed=False, label="UInt16"), "sa_data": SimTypeFixedSizeArray(SimTypeBottom(label="CHAR"), 14)}, name="SOCKADDR", pack=False, align=None), offset=0), offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0)], SimTypeBottom(label="Void"), arg_names=["lpOutputBuffer", "dwReceiveDataLength", "dwLocalAddressLength", "dwRemoteAddressLength", "LocalSockaddr", "LocalSockaddrLength", "RemoteSockaddr", "RemoteSockaddrLength"]),
        #
        'EnumProtocolsA': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["lpiProtocols", "lpProtocolBuffer", "lpdwBufferLength"]),
        #
        'EnumProtocolsW': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["lpiProtocols", "lpProtocolBuffer", "lpdwBufferLength"]),
        #
        'GetAddressByNameA': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="Guid"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"lpServiceCallbackProc": SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeBottom(label="Void"), arg_names=["lParam", "hAsyncTaskHandle"]), offset=0), "lParam": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "hAsyncTaskHandle": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)}, name="SERVICE_ASYNC_INFO", pack=False, align=None), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), label="LPArray", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["dwNameSpace", "lpServiceType", "lpServiceName", "lpiProtocols", "dwResolution", "lpServiceAsyncInfo", "lpCsaddrBuffer", "lpdwBufferLength", "lpAliasBuffer", "lpdwAliasBufferLength"]),
        #
        'GetAddressByNameW': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="Guid"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"lpServiceCallbackProc": SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeBottom(label="Void"), arg_names=["lParam", "hAsyncTaskHandle"]), offset=0), "lParam": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "hAsyncTaskHandle": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)}, name="SERVICE_ASYNC_INFO", pack=False, align=None), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["dwNameSpace", "lpServiceType", "lpServiceName", "lpiProtocols", "dwResolution", "lpServiceAsyncInfo", "lpCsaddrBuffer", "lpdwBufferLength", "lpAliasBuffer", "lpdwAliasBufferLength"]),
        #
        'GetTypeByNameA': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeBottom(label="Guid"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["lpServiceName", "lpServiceType"]),
        #
        'GetTypeByNameW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeBottom(label="Guid"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["lpServiceName", "lpServiceType"]),
        #
        'GetNameByTypeA': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Guid"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["lpServiceType", "lpServiceName", "dwNameLength"]),
        #
        'GetNameByTypeW': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Guid"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["lpServiceType", "lpServiceName", "dwNameLength"]),
        #
        'SetServiceA': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="SET_SERVICE_OPERATION"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"lpServiceType": SimTypePointer(SimTypeBottom(label="Guid"), offset=0), "lpServiceName": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "lpComment": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "lpLocale": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "dwDisplayHint": SimTypeInt(signed=False, label="RESOURCE_DISPLAY_TYPE"), "dwVersion": SimTypeInt(signed=False, label="UInt32"), "dwTime": SimTypeInt(signed=False, label="UInt32"), "lpMachineName": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "lpServiceAddress": SimTypePointer(SimStruct({"dwAddressCount": SimTypeInt(signed=False, label="UInt32"), "Addresses": SimTypePointer(SimStruct({"dwAddressType": SimTypeInt(signed=False, label="UInt32"), "dwAddressFlags": SimTypeInt(signed=False, label="UInt32"), "dwAddressLength": SimTypeInt(signed=False, label="UInt32"), "dwPrincipalLength": SimTypeInt(signed=False, label="UInt32"), "lpAddress": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "lpPrincipal": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="SERVICE_ADDRESS", pack=False, align=None), offset=0)}, name="SERVICE_ADDRESSES", pack=False, align=None), offset=0), "ServiceSpecificInfo": SimTypeBottom(label="BLOB")}, name="SERVICE_INFOA", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"lpServiceCallbackProc": SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeBottom(label="Void"), arg_names=["lParam", "hAsyncTaskHandle"]), offset=0), "lParam": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "hAsyncTaskHandle": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)}, name="SERVICE_ASYNC_INFO", pack=False, align=None), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["dwNameSpace", "dwOperation", "dwFlags", "lpServiceInfo", "lpServiceAsyncInfo", "lpdwStatusFlags"]),
        #
        'SetServiceW': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="SET_SERVICE_OPERATION"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"lpServiceType": SimTypePointer(SimTypeBottom(label="Guid"), offset=0), "lpServiceName": SimTypePointer(SimTypeChar(label="Char"), offset=0), "lpComment": SimTypePointer(SimTypeChar(label="Char"), offset=0), "lpLocale": SimTypePointer(SimTypeChar(label="Char"), offset=0), "dwDisplayHint": SimTypeInt(signed=False, label="RESOURCE_DISPLAY_TYPE"), "dwVersion": SimTypeInt(signed=False, label="UInt32"), "dwTime": SimTypeInt(signed=False, label="UInt32"), "lpMachineName": SimTypePointer(SimTypeChar(label="Char"), offset=0), "lpServiceAddress": SimTypePointer(SimStruct({"dwAddressCount": SimTypeInt(signed=False, label="UInt32"), "Addresses": SimTypePointer(SimStruct({"dwAddressType": SimTypeInt(signed=False, label="UInt32"), "dwAddressFlags": SimTypeInt(signed=False, label="UInt32"), "dwAddressLength": SimTypeInt(signed=False, label="UInt32"), "dwPrincipalLength": SimTypeInt(signed=False, label="UInt32"), "lpAddress": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "lpPrincipal": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="SERVICE_ADDRESS", pack=False, align=None), offset=0)}, name="SERVICE_ADDRESSES", pack=False, align=None), offset=0), "ServiceSpecificInfo": SimTypeBottom(label="BLOB")}, name="SERVICE_INFOW", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"lpServiceCallbackProc": SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeBottom(label="Void"), arg_names=["lParam", "hAsyncTaskHandle"]), offset=0), "lParam": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "hAsyncTaskHandle": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)}, name="SERVICE_ASYNC_INFO", pack=False, align=None), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["dwNameSpace", "dwOperation", "dwFlags", "lpServiceInfo", "lpServiceAsyncInfo", "lpdwStatusFlags"]),
        #
        'GetServiceA': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="Guid"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimStruct({"lpServiceCallbackProc": SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeBottom(label="Void"), arg_names=["lParam", "hAsyncTaskHandle"]), offset=0), "lParam": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "hAsyncTaskHandle": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)}, name="SERVICE_ASYNC_INFO", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["dwNameSpace", "lpGuid", "lpServiceName", "dwProperties", "lpBuffer", "lpdwBufferSize", "lpServiceAsyncInfo"]),
        #
        'GetServiceW': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="Guid"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimStruct({"lpServiceCallbackProc": SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeBottom(label="Void"), arg_names=["lParam", "hAsyncTaskHandle"]), offset=0), "lParam": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "hAsyncTaskHandle": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)}, name="SERVICE_ASYNC_INFO", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["dwNameSpace", "lpGuid", "lpServiceName", "dwProperties", "lpBuffer", "lpdwBufferSize", "lpServiceAsyncInfo"]),
    }

lib.set_prototypes(prototypes)



from typing import Union, Optional
from enum import Enum, auto

import claripy
import ailment
from archinfo import Arch, RegisterOffset

from ...calling_conventions import SimFunctionArgument, SimRegArg, SimStackArg
from ...engines.light import SpOffset
from .heap_address import HeapAddress


class AtomKind(Enum):
    """
    An enum indicating the class of an atom
    """

    REGISTER = auto()
    MEMORY = auto()
    TMP = auto()
    GUARD = auto()
    CONSTANT = auto()


class Atom:
    """
    This class represents a data storage location manipulated by IR instructions.

    It could either be a Tmp (temporary variable), a Register, a MemoryLocation.
    """

    __slots__ = ("_hash", "size")

    def __init__(self, size):
        """
        :param size:  The size of the atom in bytes
        """
        self.size = size
        self._hash = None

    def __repr__(self):
        raise NotImplementedError()

    @property
    def bits(self) -> int:
        return self.size * 8

    @property
    def _size(self):
        return self.size

    @_size.setter
    def _size(self, v):
        self.size = v

    @staticmethod
    def from_ail_expr(expr: ailment.Expr.Expression, arch: Arch, full_reg: bool = False) -> "Register":
        if isinstance(expr, ailment.Expr.Register):
            if full_reg:
                reg_name = arch.translate_register_name(expr.reg_offset)
                return Register(arch.registers[reg_name][0], arch.registers[reg_name][1], arch)
            else:
                return Register(expr.reg_offset, expr.size, arch)
        raise TypeError(f"Expression type {type(expr)} is not yet supported")

    @staticmethod
    def from_argument(
        argument: SimFunctionArgument, arch: Arch, full_reg=False, sp: Optional[int] = None
    ) -> Union["Register", "MemoryLocation"]:
        """
        Instanciate an `Atom` from a given argument.

        :param argument: The argument to create a new atom from.
        :param registers: A mapping representing the registers of a given architecture.
        :param full_reg: Whether to return an atom indicating the entire register if the argument only specifies a
                        slice of the register.
        :param sp:      The current stack offset. Optional. Only used when argument is a SimStackArg.
        """
        if isinstance(argument, SimRegArg):
            if full_reg:
                return Register(arch.registers[argument.reg_name][0], arch.registers[argument.reg_name][1], arch)
            else:
                return Register(arch.registers[argument.reg_name][0] + argument.reg_offset, argument.size, arch)
        elif isinstance(argument, SimStackArg):
            if sp is None:
                raise ValueError("You must provide a stack pointer to translate a SimStackArg")
            return MemoryLocation(SpOffset(arch.bits, argument.stack_offset + sp), argument.size)
        else:
            raise TypeError("Argument type %s is not yet supported." % type(argument))

    @staticmethod
    def reg(thing: Union[str, RegisterOffset], size: Optional[int] = None, arch: Optional[Arch] = None) -> "Register":
        """
        Create a Register atom.

        :param thing:   The register offset (e.g., project.arch.registers["rax"][0]) or the register name (e.g., "rax").
        :param size:    Size of the register atom. Must be provided when creating the atom using a register offset.
        :param arch:    The architecture. Must be provided when creating the atom using a register name.
        :return:        The Register Atom object.
        """

        if isinstance(thing, str):
            if arch is None:
                raise ValueError(
                    "Cannot create a Register Atom by register name without having an architecture "
                    "specified through arch!"
                )
            if thing not in arch.registers:
                raise ValueError(f"Unknown register name {thing} for architecture {arch.name}")
            reg_offset, size_ = arch.registers[thing]
            if size is None:
                size = size_
        elif isinstance(thing, RegisterOffset):
            reg_offset = thing
            if size is None:
                raise ValueError("You must provide a size when specifying the register offset")
        else:
            raise TypeError(
                "Unsupported type of register. It must be a string (for register name) or an int (for "
                "register offset)"
            )
        return Register(reg_offset, size, arch=arch)

    register = reg

    @staticmethod
    def mem(addr: Union[SpOffset, HeapAddress, int], size: int, endness: Optional[str] = None) -> "MemoryLocation":
        """
        Create a MemoryLocation atom,

        :param addr:        The memory location. Can be an SpOffset for stack variables, an int for global memory
                            variables, or a HeapAddress for items on the heap.
        :param size:        Size of the atom.
        :param endness:     Optional, either "Iend_LE" or "Iend_BE".
        :return:            The MemoryLocation Atom object.
        """
        return MemoryLocation(addr, size, endness=endness)

    memory = mem

    def _identity(self):
        raise NotImplementedError()

    def __hash__(self):
        if self._hash is None:
            self._hash = hash(self._identity())
        return self._hash

    def __eq__(self, other):
        return type(self) is type(other) and self._identity() == other._identity()


class GuardUse(Atom):
    """
    Implements a guard use.
    """

    __slots__ = ("target",)

    def __init__(self, target):
        super().__init__(0)
        self.target = target

    def __repr__(self):
        return "<Guard %#x>" % self.target

    def _identity(self):
        return (self.target,)


class ConstantSrc(Atom):
    """
    Represents a constant.
    """

    __slots__ = ("value",)

    def __init__(self, value: int, size: int):
        super().__init__(size)
        self.value: int = value

    def __repr__(self):
        return f"<Const {self.value}>"

    def _identity(self):
        return (self.value, self.size)


class Tmp(Atom):
    """
    Represents a variable used by the IR to store intermediate values.
    """

    __slots__ = ("tmp_idx",)

    def __init__(self, tmp_idx: int, size: int):
        super().__init__(size)
        self.tmp_idx = tmp_idx

    def __repr__(self):
        return "<Tmp %d>" % self.tmp_idx

    def _identity(self):
        return hash(("tmp", self.tmp_idx))


class Register(Atom):
    """
    Represents a given CPU register.

    As an IR abstracts the CPU design to target different architectures, registers are represented as a separated memory
    space.
    Thus a register is defined by its offset from the base of this memory and its size.

    :ivar int reg_offset:    The offset from the base to define its place in the memory bloc.
    :ivar int size:          The size, in number of bytes.
    """

    __slots__ = (
        "reg_offset",
        "arch",
    )

    def __init__(self, reg_offset: RegisterOffset, size: int, arch: Optional[Arch] = None):
        super().__init__(size)

        self.reg_offset = reg_offset
        self.arch = arch

    def __repr__(self):
        return "<Reg %s<%d>>" % (self.name, self.size)

    def _identity(self):
        return (self.reg_offset, self.size)

    @property
    def name(self) -> str:
        return (
            str(self.reg_offset) if self.arch is None else self.arch.translate_register_name(self.reg_offset, self.size)
        )


class MemoryLocation(Atom):
    """
    Represents a memory slice.

    It is characterized by its address and its size.
    """

    __slots__ = (
        "addr",
        "endness",
    )

    def __init__(self, addr: Union[SpOffset, HeapAddress, int], size: int, endness: Optional[str] = None):
        """
        :param int addr: The address of the beginning memory location slice.
        :param int size: The size of the represented memory location, in bytes.
        """
        super().__init__(size)

        self.addr: Union[SpOffset, int, claripy.ast.BV] = addr
        self.endness = endness

    def __repr__(self):
        address_format = hex(self.addr) if type(self.addr) is int else self.addr
        stack_format = " (stack)" if self.is_on_stack else ""
        size = "%d" % self.size if isinstance(self.size, int) else self.size

        return f"<Mem {address_format}<{size}>{stack_format}>"

    @property
    def is_on_stack(self) -> bool:
        """
        True if this memory location is located on the stack.
        """
        return isinstance(self.addr, SpOffset)

    @property
    def symbolic(self) -> bool:
        if isinstance(self.addr, int):
            return False
        elif isinstance(self.addr, SpOffset):
            return type(self.addr.offset) is not int
        return True

    def __eq__(self, other):
        # pylint:disable=isinstance-second-argument-not-valid-type
        return (
            type(other) is MemoryLocation
            and (
                self.addr is other.addr
                if (isinstance(self.addr, claripy.ast.BV) or isinstance(other.addr, claripy.ast.BV))
                else self.addr == other.addr
            )
            and self.size == other.size
            and self.endness == other.endness
        )

    __hash__ = Atom.__hash__

    def _identity(self):
        return self.addr, self.size, self.endness


atom_kind_mapping = {
    AtomKind.REGISTER: Register,
    AtomKind.MEMORY: MemoryLocation,
    AtomKind.TMP: Tmp,
    AtomKind.GUARD: GuardUse,
    AtomKind.CONSTANT: ConstantSrc,
}



from . import JNISimProcedure

# pylint: disable=arguments-differ,unused-argument


class GetVersion(JNISimProcedure):
    return_ty = "int"

    def run(self, ptr_env):
        # return JNI version 1.8
        return 0x00010008



from angr.procedures.stubs.format_parser import ScanfFormatParser

from cle.backends.externs.simdata.io_file import io_file_data_for_arch


class fscanf(ScanfFormatParser):
    # pylint:disable=arguments-differ

    def run(self, file_ptr, fmt):  # pylint:disable=unused-argument
        # TODO handle errors

        fd_offset = io_file_data_for_arch(self.state.arch)["fd"]
        fd = self.state.mem[file_ptr + fd_offset :].int.resolved
        simfd = self.state.posix.get_fd(fd)
        if simfd is None:
            return -1

        fmt_str = self._parse(fmt)
        items = fmt_str.interpret(self.va_arg, simfd=simfd)
        return items



from . import ExplorationTechnique
import psutil


class MemoryWatcher(ExplorationTechnique):
    """Memory Watcher

    Args:
        min_memory (int,optional): Minimum amount of free memory in MB before
                    stopping execution (default: 95% memory use)
        memory_stash (str, optional): What to call the low memory stash
                    (default: 'lowmem')

    At each step, keep an eye on how much memory is left on the system. Stash
    off states to effectively stop execution if we're below a given threshold.
    """

    def __init__(self, min_memory=512, memory_stash="lowmem"):
        super().__init__()

        if min_memory is not None:
            self.min_memory = 1024 * 1024 * min_memory

        else:
            self.min_memory = int(psutil.virtual_memory().total * 0.05)

        self.memory_stash = memory_stash

    def setup(self, simgr):
        if self.memory_stash not in simgr.stashes:
            simgr.stashes[self.memory_stash] = []

    def step(self, simgr, stash="active", **kwargs):
        if psutil.virtual_memory().available <= self.min_memory:
            simgr.move(from_stash="active", to_stash=self.memory_stash)

        else:
            simgr = simgr.step(stash=stash, **kwargs)

        return simgr



import claripy
import logging

from ..java import JavaSimProcedure
from ...engines.soot.values import SimSootValue_ThisRef

log = logging.getLogger(name=__name__)

ELEMS = "elems"
SIZE = "size"
INDEX = "index"


class GetIterator(JavaSimProcedure):
    __provides__ = (
        ("java.util.Collection", "iterator()"),
        ("java.util.Set", "iterator()"),
        ("java.util.List", "iterator()"),
        ("java.util.LinkedList", "iterator()"),
        ("java.util.AbstractSequentialList", "iterator()"),
    )

    def run(self, this_ref):
        log.debug(f"Called SimProcedure java.util.*.iterator with args: {this_ref}")

        iterator_ref = SimSootValue_ThisRef.new_object(self.state, "java.util.Iterator")
        elems_array_ref = this_ref.get_field(self.state, ELEMS, "java.lang.Object[]")
        iterator_ref.store_field(self.state, ELEMS, "java.lang.Object[]", elems_array_ref)
        collection_size = this_ref.get_field(self.state, SIZE, "int")
        iterator_size = claripy.BVV(self.state.solver.eval(collection_size), 32)
        iterator_ref.store_field(self.state, SIZE, "int", iterator_size)
        iterator_ref.store_field(self.state, INDEX, "int", claripy.BVV(0, 32))

        return iterator_ref



import logging

from archinfo import ArchSoot

from .base import SimSootExpr
from ..values.thisref import SimSootValue_ThisRef

l = logging.getLogger("angr.engines.soot.expressions.cast")


class SimSootExpr_Cast(SimSootExpr):
    def _execute(self):
        # get value
        local = self._translate_value(self.expr.value)
        value_uncasted = self.state.memory.load(local)
        # cast value
        if self.expr.cast_type in ArchSoot.primitive_types:
            javavm_simos = self.state.project.simos
            self.expr = javavm_simos.cast_primitive(self.state, value_uncasted, to_type=self.expr.cast_type)
        else:
            self.expr = SimSootValue_ThisRef(heap_alloc_id=value_uncasted.heap_alloc_id, type_=self.expr.cast_type)



import logging

from .plugin import SimStatePlugin

l = logging.getLogger(name=__name__)


class SimStateGlobals(SimStatePlugin):
    def __init__(self, backer=None):
        super().__init__()
        self._backer = backer if backer is not None else {}

    def set_state(self, state):
        pass

    def merge(self, others, merge_conditions, common_ancestor=None):  # pylint: disable=unused-argument
        for other in others:
            for k in other.keys():
                if k not in self:
                    self[k] = other[k]

        return True

    def widen(self, others):  # pylint: disable=unused-argument
        l.warning("Widening is unimplemented for globals")
        return False

    def __getitem__(self, k):
        return self._backer[k]

    def __setitem__(self, k, v):
        self._backer[k] = v

    def __delitem__(self, k):
        del self._backer[k]

    def __contains__(self, k):
        return k in self._backer

    def keys(self):
        return self._backer.keys()

    def values(self):
        return self._backer.values()

    def items(self):
        return self._backer.items()

    def get(self, k, alt=None):
        return self._backer.get(k, alt)

    def pop(self, k, alt=None):
        return self._backer.pop(k, alt)

    @SimStatePlugin.memo
    def copy(self, memo):  # pylint: disable=unused-argument
        return SimStateGlobals(dict(self._backer))


from angr.sim_state import SimState

SimState.register_default("globals", SimStateGlobals)
